# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for parsing complex query strings into Django ORM Q objects.
Provides functionality to handle logical operators (AND, OR), negations,
and nested conditions.

Tags: RAG, EXPORT
"""

import copy
import dataclasses
import operator
from enum import Enum
from functools import reduce
from hashlib import md5
from itertools import chain
from typing import Union
from urllib.parse import unquote, urlencode

from django.contrib.gis.db.models import PointField
from django.contrib.gis.geos import GEOSGeometry, Polygon
from django.contrib.postgres.fields import ArrayField, RangeField
from django.db import models
from django.db.models import Exists, Q, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.utils.dateparse import parse_date, parse_datetime

from bazis.core.models_abstract import InitialBase
from bazis.core.utils.functools import get_attr
from bazis.core.utils.model_meta import FieldsInfo
from bazis.core.utils.orm import apply_calc_queryset, calc_cached_property


OP_NEG = '~'
SEARCH_TERM = '$search'
RANGE_SUFFIX = {'gt', 'gte', 'lt', 'lte'}
BOOL_SUFFIX = {'isnull'}
BOOLS = {
    'true': True,
    '1': True,
    '': False,
    'false': False,
    '0': False,
    True: True,
    False: False,
}
BOOLS_NEG = {k for k, v in BOOLS.items() if v is False}
NULL = 'null'

LOOKUP_PREFIXES: dict[str, str] = {
    '^': 'istartswith',
    '=': 'iexact',
    '@': 'search',
    '$': 'iregex',
}

DJANGO_SEARCH_FIELDS = [
    models.CharField,
    models.TextField,
    models.IntegerField,
]


class QueryOp(Enum):
    AND = 'and'
    OR = 'or'


@dataclasses.dataclass
class FieldDummy:
    name: str
    model: type[models.Model] = None
    related_model: type[models.Model] = None


# QueryComplexItem = namedtuple('QueryComplexItem', ['key', 'value', 'is_neg'])


class QueryComplexExpression:
    """Represents a set of QueryComplexItems, logical operators (AND, OR) and QueryComplexExpression."""

    is_neg: bool
    conditions_set: list['QueryComplexItem | QueryComplexExpression | QueryOp']

    def __init__(
        self,
        conditions_set: list['QueryComplexItem | QueryComplexExpression | QueryOp'],
        is_neg: bool = False,
    ) -> None:
        """
        Initializes a QueryComplexExpression with a set of conditions.

        :param conditions_set: A set of QueryComplexItems, QueryComplexExpression, or QueryOps.
        :param is_neg: Indicates whether the condition set is negated.
        """
        self.is_neg = is_neg
        self.conditions_set = conditions_set

    def __str__(self) -> str:
        """
        Returns a string representation of the conditions set.

        :return: A formatted string containing the conditions.
        """
        return f'QueryComplexExpression({self.conditions_set}, is_neg={self.is_neg})'

    __repr__ = __str__


class QueryComplexItem:
    """Represents a single query condition consisting of a key-value pair."""

    is_neg: bool
    key: str
    value: str
    parent: 'QueryComplex | None'

    def __init__(
        self, key: str, value: str, is_neg: bool = False, parent: 'QueryComplex | None' = None
    ) -> None:
        """
        Initializes a QueryComplexItem with a key, value, and optional negation.

        :param is_neg: Indicates whether the condition is negated.
        :param key: The field name of the query condition.
        :param value: The corresponding value of the condition.
        :param parent: The parent QueryComplex instance, if applicable.
        """
        self.is_neg = is_neg
        self.key = key
        self.value = value
        self.parent = parent

    def __hash__(self):
        return int(md5(str(self).encode()).hexdigest(), 16)

    def __eq__(self, other):
        return type(self) is type(other) and hash(self) == hash(other)

    def __str__(self):
        return urlencode({self.key: self.value}) + f'?{self.is_neg}'

    def __deepcopy__(self, memo):
        return QueryComplexItem(key=self.key, value=self.value, is_neg=self.is_neg)

    __repr__ = __str__

    def delete(self):
        self.parent.remove(self)

    def replace(self, key: str, value: str, is_neg: bool = False):
        self.key = key
        self.value = value
        self.is_neg = is_neg
        self.parent.rebalanced()

    def add_node(self, node: Union[str, dict, 'QueryComplex'], op: QueryOp):
        if isinstance(node, str | dict):
            node = QueryComplex.from_data(node)

        # if node has only one condition
        if not node.right:
            node = node.left

        current_parent = self.parent
        node_union = QueryComplex(left=self, right=node, op=op, parent=current_parent)

        if current_parent.left == self:
            current_parent.set_left(node_union)
        else:
            current_parent.set_right(node_union)

        next_node = current_parent
        while parent := next_node.parent:
            parent.rebalanced()
            next_node = parent
        return node_union

    def __and__(self, new: Union[str, dict, 'QueryComplexItem', 'QueryComplex']):
        """
        Overloads the & operator to add a condition using 'and'.
        """
        return self.add_node(new, QueryOp.AND)

    def __or__(self, new: Union[str, dict, 'QueryComplexItem', 'QueryComplex']):
        """
        Overloads the | operator to add a condition using 'or'.
        """
        return self.add_node(new, QueryOp.OR)


class QueryComplex:
    parent: Union['QueryComplex', type[None]] = None
    right: Union['QueryComplex', QueryComplexItem, type[None]] = None
    op: QueryOp | type[None] = None
    left: Union['QueryComplex', QueryComplexItem, type[None]] = None
    is_neg: bool = False

    _hash = None
    _dict = {}

    @classmethod
    def _parse_query(cls, query: str) -> QueryComplexExpression:
        """
        Parses the query string containing conditions with logical operators and nesting,
        and converts it into a QueryComplexExpression object.
        """
        query = query.replace('[', '(').replace(']', ')')
        query = query.replace('|&', '|').replace('&|', '|')

        struct = {
            'conditions': QueryComplexExpression(conditions_set=[], is_neg=False),
            'stack': [],
            'key': '',
            'value': None,
            'is_neg': False,
        }
        struct['stack'].append(struct['conditions'])

        def add_cond() -> None:
            """
            If the key and value variables are set, creates a QueryComplexItem with the given is_neg flag
            and appends it to conditions.conditions_set.
            """
            if struct['key'] and struct['value'] is not None:
                struct['conditions'].conditions_set.append(
                    QueryComplexItem(
                        key=struct['key'].strip(),
                        value=unquote(struct['value']).strip().replace('"', '').replace("'", ''),
                        is_neg=struct['is_neg'],
                    )
                )
            struct['key'] = ''
            struct['value'] = None
            struct['is_neg'] = False

        def add_op(op) -> None:
            """
            Appends the operator (AND/OR).
            """
            struct['conditions'].conditions_set.append(op)

        def add_neg() -> None:
            """
            Sets is_neg to True, indicating that the next condition should be negated.
            """
            struct['is_neg'] = True

        def open_nested() -> None:
            """
            Creates a new QueryComplexExpression for nested conditions and pushes it onto the stack.
            """
            struct['conditions'] = QueryComplexExpression(
                conditions_set=[], is_neg=struct['is_neg']
            )
            struct['stack'].append(struct['conditions'])

        def close_nested() -> None:
            """
            Pops the last nested level from the stack and appends it to the parent's conditions_set.
            """
            nested = struct['stack'].pop()
            struct['conditions'] = struct['stack'][-1]
            struct['conditions'].conditions_set.append(nested)

        for ch in query:
            if ch == '&':
                add_cond()
                add_op(QueryOp.AND)
            elif ch == '|':
                add_cond()
                add_op(QueryOp.OR)
            elif ch == OP_NEG:
                add_neg()
            elif ch == '(':
                open_nested()
                add_cond()
            elif ch == ')':
                add_cond()
                close_nested()
            elif ch == '=':
                struct['value'] = ''
            elif struct['value'] is not None:
                struct['value'] += ch
            else:
                struct['key'] += ch
        else:
            add_cond()

        return struct['conditions']

    @classmethod
    def from_data(cls, data: str | dict) -> Union['QueryComplex', QueryComplexItem]:
        def get_condition_pairs(conditions: QueryComplexExpression) -> QueryComplexExpression:
            """Splits a list of conditions into nested condition pairs for further processing in QueryComplex."""
            pair = []

            for condition in conditions.conditions_set:
                if isinstance(condition, QueryComplexExpression):
                    condition = get_condition_pairs(condition)

                if len(pair) < 3:
                    pair.append(condition)
                else:
                    pair = [pair, condition]

            return QueryComplexExpression(conditions_set=pair, is_neg=conditions.is_neg)

        def make_query_complex(
            conditions: QueryComplexExpression,
        ) -> QueryComplex | QueryComplexItem | None:
            """Constructs a QueryComplex object from a conditions_hierarchy of conditions."""
            if isinstance(conditions, QueryComplexItem) or conditions is None:
                return conditions
            elif isinstance(conditions, QueryComplexExpression):
                is_neg = conditions.is_neg
                conditions_set = conditions.conditions_set
            elif isinstance(conditions, list):
                is_neg = False
                conditions_set = conditions

            left, op, right = (conditions_set + [None] * 3)[:3]
            assert op is None or op in QueryOp

            left = make_query_complex(left)
            right = make_query_complex(right)

            return QueryComplex(left=left, right=right, op=op, is_neg=is_neg)

        if isinstance(data, dict):
            data = {k: ','.join(v) if isinstance(v, list | tuple) else v for k, v in data.items()}
            data = urlencode(data)

        parsed_conditions = cls._parse_query(data)
        splited_conditions = get_condition_pairs(conditions=parsed_conditions)
        query_complex = make_query_complex(conditions=splited_conditions)
        return query_complex

    def __init__(
        self,
        left: Union['QueryComplex', QueryComplexItem] = None,
        right: Union['QueryComplex', QueryComplexItem] = None,
        op: QueryOp = None,
        parent: 'QueryComplex' = None,
        is_neg=False,
    ):
        self.left = left
        self.right = right
        self.op = op
        self.parent = parent
        self.is_neg = is_neg
        self.rebalanced()

    def __hash__(self):
        return self._hash or 0

    def __bool__(self):
        return bool(self.left or self.right)

    def __deepcopy__(self, memo):
        def _query_perform(query: QueryComplex, parent: QueryComplex | None):
            left = query.left
            right = query.right
            return QueryComplex(
                left=(
                    _query_perform(left, query)
                    if isinstance(left, QueryComplex)
                    else copy.deepcopy(left)
                ),
                op=query.op,
                right=(
                    _query_perform(right, query)
                    if isinstance(right, QueryComplex)
                    else copy.deepcopy(right)
                ),
                parent=parent,
                is_neg=query.is_neg,
            )

        return _query_perform(self, None)

    def rebalanced(self):
        if not self.left and not self.right:
            self.op = None
            self.parent = None
            self.is_neg = False
            self._hash = None
            return

        if self.left and self.right:
            pair = sorted([self.left, self.right], key=lambda x: hash(x))
        elif self.left:
            pair = [self.left, None]
            self.op = None
        else:
            pair = [self.right, None]
            self.op = None

        self.left = pair[0]
        self.right = pair[1]

        if self.left:
            self.left.parent = self
        if self.right:
            self.right.parent = self

        left_hash = '0' if self.left is None else hash(self.left)
        op_value = '' if self.op is None else self.op.value
        right_hash = '0' if self.right is None else hash(self.right)

        self._hash = int(md5(f'{left_hash}.{op_value}.{right_hash}'.encode()).hexdigest(), 16)

    def set_left(self, node: Union['QueryComplex', QueryComplexItem]):
        self.left = node
        self.rebalanced()

    def set_right(self, node: Union['QueryComplex', QueryComplexItem]):
        self.right = node
        self.rebalanced()

    def add_node(self, node: Union[str, dict, 'QueryComplex'], op: QueryOp):
        if isinstance(node, str | dict):
            node = QueryComplex.from_data(node)

        # for cases when the structure is either empty, has only one element, or has no parent
        if not self:
            self.left = node.left
            self.right = node.right
            self.op = node.op
            self.is_neg = node.is_neg
            self.rebalanced()
            return self
        elif not self.right:
            self.right = node
            self.op = op
            self.rebalanced()
            return self
        elif not self.parent:
            self.left = QueryComplex(
                left=self.left, right=self.right, op=self.op, is_neg=self.is_neg, parent=self
            )
            self.right = node
            self.op = op
            self.is_neg = node.is_neg
            self.rebalanced()
            return self

        # if node has only one condition
        if not node.right:
            node = node.left

        current_parent = self.parent
        node_union = QueryComplex(left=self, right=node, op=op, parent=current_parent)

        if current_parent.left == self:
            current_parent.set_left(node_union)
        else:
            current_parent.set_right(node_union)

        next_node = current_parent
        while parent := next_node.parent:
            parent.rebalanced()
            next_node = parent
        return node_union

    def __and__(self, new: Union[str, dict, QueryComplexItem, 'QueryComplex']):
        """
        Overloads the & operator to add a condition using 'and'.
        """
        obj = copy.deepcopy(self)
        return obj.add_node(new, QueryOp.AND)

    def __iadd__(self, new: Union[str, dict, QueryComplexItem, 'QueryComplex']):
        """
        Overloads the += operator to add a condition using 'and'.
        """
        return self.add_node(new, QueryOp.AND)

    def __or__(self, new: Union[str, dict, QueryComplexItem, 'QueryComplex']):
        """
        Overloads the | operator to add a condition using 'or'.
        """
        obj = copy.deepcopy(self)
        return obj.add_node(new, QueryOp.OR)

    def __ior__(self, new: Union[str, dict, QueryComplexItem, 'QueryComplex']):
        """
        Overloads the |= operator to add a condition using 'or'.
        """
        return self.add_node(new, QueryOp.OR)

    def remove(self, node: Union[QueryComplexItem, 'QueryComplex']):
        """
        Removes a condition from the internal structure.
        """
        if hash(node) not in (hash(self.left), hash(self.right)):
            raise ValueError('Node not found in the structure.')

        if hash(self.left) == hash(node):
            alive = self.right
        else:
            alive = self.left

        if self.parent:
            if self == self.parent.left:
                self.parent.set_left(alive)
            else:
                self.parent.set_right(alive)

            next_node = self.parent
            while parent := next_node.parent:
                parent.rebalanced()
                next_node = parent
        elif alive:
            self.left = alive
            self.right = None
            self.rebalanced()
        else:
            self.left = None
            self.right = None
            self.rebalanced()

        return alive or self

    def dump(self):
        """
        Converts the internal structure to a dictionary for easier visualization.
        """

        def _query_perform(query: QueryComplex):
            left = query.left
            right = query.right
            return [
                _query_perform(left) if isinstance(left, QueryComplex) else left,
                query.op.value if query.op else None,
                _query_perform(right) if isinstance(right, QueryComplex) else right,
            ]

        return _query_perform(self)

    # def _to_dict(self):
    #     """
    #     Transforms the structure to the desired dictionary format.
    #     """
    #     result = defaultdict(list)
    #
    #     def transform_structure(conditions=None, prefix=None, op=QueryOp.AND):
    #         if conditions is None:
    #             conditions = self
    #
    #         nested_count = 0
    #
    #         for cond in conditions:
    #             if isinstance(cond, QueryComplexItem):
    #                 full_key = ['', op.value]
    #                 if prefix:
    #                     full_key.append(prefix)
    #                 full_key.append(cond.key)
    #                 if cond.is_neg:
    #                     full_key.append('is_neg')
    #                 result['__'.join(full_key)].append(cond.value)
    #             elif cond in QueryOp:
    #                 op = cond
    #             elif isinstance(cond, QueryComplex):
    #                 nested_prefix = []
    #                 if prefix:
    #                     nested_prefix.append(prefix)
    #                 nested_prefix.append(f'{nested_count}')
    #                 nested_count += 1
    #                 transform_structure(cond, '__'.join(nested_prefix), op)
    #
    #     transform_structure()
    #     self._dict = result

    # def __getitem__(self, item):
    #     return self._dict[item]
    #
    # def keys(self):
    #     return self._dict.keys()


class QueryToOrm:
    q: Q
    fields_calc: list

    @classmethod
    def qs_apply(
        cls,
        queryset: QuerySet,
        query_complex: QueryComplex | str,
        filters_aliases: dict[str, str] = None,
        fiter_context: dict = None,
    ):
        """
        Applies the filtering logic to the given QuerySet based on the provided query string,
        filter aliases, and context.

        :param queryset: The Django QuerySet to filter.
        :param query_complex: The query string to parse and apply as filters.
        :param filters_aliases: Optional dictionary mapping filter aliases to actual field names.
        :param fiter_context: Optional context dictionary for additional filtering logic.
        :return: Filtered QuerySet.
        """
        if not query_complex:
            return queryset

        q_orm = cls(query_complex, queryset.model, filters_aliases, fiter_context)
        if q_orm.fields_calc:
            queryset = apply_calc_queryset(queryset, q_orm.fields_calc, context=fiter_context)
        queryset = queryset.filter(q_orm.q)
        return queryset

    def __init__(
        self,
        query_complex: QueryComplex | str,
        model: type[models.Model],
        filters_aliases: dict[str, str] = None,
        fiter_context: dict = None,
    ):
        """
        Initializes the Filtering instance with the model, query string, filter aliases, and context.

        :param model: The Django model to filter.
        :param query_str: The query string to parse and apply as filters.
        :param filters_aliases: Optional dictionary mapping filter aliases to actual field names.
        :param fiter_context: Optional context dictionary for additional filtering logic.
        """
        if isinstance(query_complex, str):
            query_complex = QueryComplex.from_data(query_complex)

        self.model = model
        self.query_complex = query_complex
        self.filters_aliases = filters_aliases or {}
        self.fiter_context = fiter_context or {}
        self.q, self.fields_calc = self._query_perform(self.query_complex)

    def _query_perform(self, query: QueryComplex):
        fields_calc = []
        q_result = []
        q = Q()

        if not query:
            return q, fields_calc

        for node in (query.left, query.right):
            if not node:
                q_result.append(Q())
                continue

            if isinstance(node, QueryComplexItem):
                q, sub_fields_calc = self._lookup_parse(node.key, node.value)
                fields_calc.extend(sub_fields_calc)
            else:
                q, sub_fields_calc = self._query_perform(node)
                fields_calc.extend(sub_fields_calc)

            if node.is_neg:
                q = ~q

            q_result.append(q)

        if query.op == QueryOp.AND:
            q = q_result[0] & q_result[1]
        else:
            q = q_result[0] | q_result[1]

        return q, fields_calc

    def _lookup_parse(self, key: str, value: str):
        """
        Parses the lookup key and value, and applies the appropriate filters to the model.

        :param key: The lookup key (e.g., 'author__name').
        :param value: The value to filter by.
        :return: A tuple containing the Q object and fields calculation list.
        """
        params = key.split(LOOKUP_SEP)
        params = list(chain(*[self.filters_aliases.get(p, p).split(LOOKUP_SEP) for p in params]))
        return self._filters_apply(params, value, self.model)

    def _filters_apply(self, params, value, model):
        """
        Applies the filters based on the parsed parameters and value to the specified model.

        :param params: List of parameters derived from the lookup key.
        :param value: The value to filter by.
        :param model: The Django model to filter.
        :return: A tuple containing the Q object and fields calculation list.
        """
        queries = [Q(), []]

        fields_info = FieldsInfo.get_fields_info(model)
        param, *params = params

        def queries_add(_q, _f=None):
            """
            Adds the given Q object and fields to the current queries list.

            :param _q: The Q object to add.
            :param _f: The fields to add.
            """
            queries[0] &= _q
            if _f:
                queries[1].extend(_f)

        if '.' in param:
            q = self._filter_by_type(param, value, model)
            if q:
                queries_add(q)

        if field_info := fields_info.fields.get(param):
            if params and param in fields_info.relations:
                queries_add(self._filters_apply_relation(params, value, field_info))
            else:
                queries_add(self._filters_apply_native(params, value, field_info))
        elif param == SEARCH_TERM and value:
            queries_add(self._filters_apply_search(model, value))
        elif func_calc := get_attr(model, param):
            if isinstance(func_calc, calc_cached_property) and func_calc.as_filter:
                if func_calc.response_type is bool or value in ('true', 'false'):
                    value = BOOLS.get(value)

                type_calc = func_calc.filter_field or FieldDummy

                type_obj = type_calc(name=param)
                type_obj.model = model

                queries_add(
                    self._filters_apply_native(params, value, type_obj), func_calc.fields_calc
                )

        return queries

    def _filter_by_type(self, model_label, ids, target_model):
        source_model = InitialBase.get_model_by_label(model_label)
        if not source_model:
            return None

        relations_by_model = source_model.get_fields_info().relations_by_model
        if target_model not in relations_by_model:
            return None

        ids = ids.split(',')

        q = Q()
        for rel in relations_by_model[target_model]:
            q |= Q(**{f'{rel.related_field.name}__in': ids})

        return q

    def _filters_apply_relation(self, params, value, field_info):
        """
        Applies filters to a related model field and constructs an Exists subquery.

        :param params: List of parameters derived from the lookup key.
        :param value: The value to filter by.
        :param field_info: Field information for the related model.
        :return: An Exists subquery representing the filter condition.
        """
        exists_qs = field_info.get_subqueryset()
        q, fields_calc = self._filters_apply(params, value, exists_qs.model)
        if fields_calc:
            exists_qs = apply_calc_queryset(exists_qs, fields_calc, context=self.fiter_context)

        # expression of the existence of a nested query
        exists_qs = Exists(exists_qs.filter(q).values('id'))

        # if the query contains an existence expression - handle it separately
        if 'exists' == params[0]:
            if value in BOOLS_NEG:
                exists_qs = ~exists_qs

        return exists_qs

    def _filters_apply_native(self, params, value, field):
        """
        For the final value, perform null conversion and apply the appropriate filter based on
        the field type.

        :param params: list of parameters derived from the lookup key.
        :param value: the value to filter by.
        :param field: the model field to apply the filter to.
        :return: a q object representing the filter condition.
        """
        if value == NULL or (
            value == '' and not isinstance(field, models.CharField | models.TextField)
        ):
            value = None

        if isinstance(field, PointField):
            func = self._func_geo
        elif isinstance(field, RangeField):
            func = self._func_range
        elif isinstance(field, models.BooleanField):
            func = self._func_bool
        elif isinstance(field, ArrayField):
            func = self._func_overlap
        elif isinstance(field, models.TextField):
            func = self._func_text
        else:
            func = self._func_exact

        return func(value, params, field)

    def _filters_apply_search(self, model, value, field_name=None):
        """
        Applies full-text search filters to the specified model based on the given value and
        optional field name.

        :param model: The Django model to filter.
        :param value: The search value.
        :param field_name: Optional field name to restrict the search to.
        :return: A Q object representing the search condition.
        """
        search = SearchToOrm(model, value, [field_name] if field_name else None)
        return search.q

    def _func_exact(self, value, params, field):
        """
        Applies exact match filters to the specified field based on the given value and parameters.

        :param value: The value to filter by.
        :param params: List of parameters derived from the lookup key.
        :param field: The model field to apply the filter to.
        :return: A Q object representing the filter condition.
        """
        if action := get_attr(params, 0):
            if action in RANGE_SUFFIX:
                return Q(**{f'{field.name}__{action}': value})
            if action in LOOKUP_PREFIXES.values():
                field_name = f'{field.name}__{action}'
                return self._filters_apply_search(field.model, value, field_name)
            if action == SEARCH_TERM:
                return self._filters_apply_search(field.model, value, field.name)
            if action in BOOL_SUFFIX:
                return self._func_bool(value, params, field)
        if isinstance(value, list | set | tuple):
            return Q(**{f'{field.name}__in': value})
        return Q(**{field.name: value})

    def _func_geo(self, value, params, field):
        """
        Applies geographic filters to the specified field based on the given value and parameters.

                Supported filters:
                - point=49.124,55.76480
                - point__near=49.124,55.76480[,100]
                - point__in_bbox=160.6,-55.95,-170,-25.89

                :param value: The geographic value to filter by.
                :param params: List of parameters derived from the lookup key.
                :param field: The geographic model field to apply the filter to.
                :return: A Q object representing the geographic filter condition.
        """
        # TODO: implement full geojson support, handle finding the nearest object,
        # inclusion in the object, etc.

        def point_near(lon, lat, distance=10):
            """
            Creates a geo-point and searches for objects within a specified radius from the target point.

            :param lon: longitude of the target point.
            :param lat: latitude of the target point.
            :param distance: radius in meters for the search (default is 10 meters).
            :return: a q object representing the geographic filter condition.
            """
            geo_point = GEOSGeometry(f'POINT({lon} {lat})', srid=4326)
            # search for objects within n-meters radius from the target
            return Q(**{f'{field.name}__distance_lte': (geo_point, distance)})

        def point_in_bbox(bbox):
            """
            Creates a bounding box polygon and searches for objects contained within the specified bounding box.

            :param bbox: List of coordinates defining the bounding box.
            :return: A Q object representing the geographic filter condition.
            """
            bbox_polygon = Polygon.from_bbox(bbox)
            return Q(**{f'{field.name}__contained': bbox_polygon})

        if value:
            if 'in_bbox' in params:
                return point_in_bbox(value.split(','))
            elif 'near' in params:
                lon, lat, *p = value.split(',')
                if p:
                    distance = int(p[0])
                else:
                    distance = 100
                return point_near(lon, lat, distance)
            else:
                lon, lat = value.split(',')
                return point_near(lon, lat, 10)
        return Q()

    def _func_overlap(self, value, params, field):
        """
        Applies overlap filters to the specified field based on the given value and parameters.

        :param value: The value to filter by, as a comma-separated string.
        :param params: List of parameters derived from the lookup key.
        :param field: The array model field to apply the filter to.
        :return: A Q object representing the overlap filter condition.
        """
        if isinstance(value, str):
            values = value.split(',')
        else:
            values = list(value)
        if values:
            if action := get_attr(params, 0):
                return Q(**{f'{field.name}__{action}': values})
            return Q(**{f'{field.name}__overlap': values})
        return Q()

    def _func_range(self, value, params, field):
        """
        Applies range filters to the specified field based on the given value and parameters.

        :param value: The range value to filter by, as a comma-separated string.
        :param params: List of parameters derived from the lookup key.
        :param field: The range model field to apply the filter to.
        :return: A Q object representing the range filter condition.
        """
        if value and params:
            action = params[0]
            if action in RangeField.class_lookups:
                start, end = value.split(',')

                if field.base_field.get_internal_type() == 'DateRangeField':
                    start = parse_date(start)
                    end = parse_date(end)
                elif field.base_field.get_internal_type() == 'DateTimeRangeField':
                    start = parse_datetime(start)
                    end = parse_datetime(end)

                return Q(**{f'{field.name}__{action}': (start, end)})
        return Q()

    def _func_bool(self, value, params, field):
        """
        Applies boolean filters to the specified field based on the given value and parameters.

        :param value: The boolean value to filter by.
        :param params: List of parameters derived from the lookup key.
        :param field: The boolean model field to apply the filter to.
        :return: A Q object representing the boolean filter condition.
        """
        is_true = value not in BOOLS_NEG
        if action := get_attr(params, 0):
            if action in BOOL_SUFFIX:
                return Q(**{f'{field.name}__{action}': is_true})
        return Q(**{field.name: is_true})

    def _func_text(self, value, params, field):
        """
        Applies text search filters to the specified field based on the given value and parameters.

        :param value: The text value to filter by.
        :param params: List of parameters derived from the lookup key.
        :param field: The text model field to apply the filter to.
        :return: A Q object representing the text search filter condition.
        """
        field_name = field.name
        if action := get_attr(params, 0):
            if action in LOOKUP_PREFIXES.values():
                field_name = f'{field_name}__{action}'
        return self._filters_apply_search(field.model, value, field_name)


class SearchToOrm:
    """
    Searching constructs a Q object based on the provided model, search string, and
    search fields. It is used to filter querysets according to the search criteria.
    """

    q: Q

    def __init__(self, model, search, search_fields=None):
        """
        Initializes the Searching class with the model, search string, and optional
        search fields. If no search fields are provided, it defaults to fields of types
        defined in DJANGO_SEARCH_FIELDS.
        """
        self.search = search and search.strip()
        self.search_fields = search_fields or [
            f.name for f in model._meta.get_fields() if type(f) in DJANGO_SEARCH_FIELDS
        ]
        if self.search_fields is None or not self.search:
            self.q = Q()
            return

        terms = self._get_terms()

        orm_lookups = [
            self._construct_search(str(search_field)) for search_field in self.search_fields
        ]

        conditions = []
        for term in terms:
            queries = [models.Q(**{orm_lookup: term}) for orm_lookup in orm_lookups]
            conditions.append(reduce(operator.or_, queries))

        self.q = reduce(operator.and_, conditions)

    def _construct_search(self, field_name):
        """
        Constructs the appropriate search lookup for a given field name. It handles
        field name prefixes and appends the correct lookup type.
        """
        if LOOKUP_SEP in field_name:
            return field_name
        lookup = LOOKUP_PREFIXES.get(field_name[0])
        if lookup:
            field_name = field_name[1:]
        else:
            lookup = 'icontains'
        return LOOKUP_SEP.join([field_name, lookup])

    def _get_terms(self) -> list[str]:
        """
        Splits the search string into individual search terms, removing any null
        characters and replacing commas with spaces.
        """
        search = self.search.replace('\x00', '')  # strip null characters
        search = search.replace(',', ' ')
        return search.split()


# # Example usage
# query_str = "contract__author=__selector__&(label=manager|~label=director|[slug=1&~slug=2]&[moon=1&~moon=2])"
# # query_str = "contract__author=__selector__&(label=manager|~label=director)"
# parser = QueryComplex.from_data(query_str)
# print(parser.dump())
#
# # Adding conditions
# parser &= {'f1': 'v1'}
# parser |= {'f2': 'v2'}
# parser &= {'f2': 'v2'}
#
# print(parser.dump())
#
# parser.left &= {'f3': 'v3'}
# print(parser.dump())



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bazis.core.routing import BazisRouter

from . import routes


router = BazisRouter(tags=['Dynamic'])

router.register(routes.DynamicVehicleModelRouteSet.as_router())
router.register(routes.DynamicVehicleBrandRouteSet.as_router())
router.register(routes.CountryRouteSet.as_router())
router.register(routes.CarrierTaskRouteSet.as_router())
router.register(routes.DivisionJsonRouteSet.as_router())
router.register(routes.DriverRouteSet.as_router())

router_is_exists_one_relation = BazisRouter(tags=['Dynamic_IsExistsOneRelation'])
router_is_exists_one_relation.register(routes.IsExistsOneRelationRouteSet.as_router())

router_is_exists_hierarchy = BazisRouter(tags=['Dynamic_IsExistsHierarchy'])
router_is_exists_hierarchy.register(routes.IsExistsHierarchyRouteSet.as_router())

router_json_one_relation = BazisRouter(tags=['Dynamic_JsonOneRelation'])
router_json_one_relation.register(routes.JsonOneRelationRouteSet.as_router())

router_json_hierarchy = BazisRouter(tags=['Dynamic_JsonHierarchy'])
router_json_hierarchy.register(routes.JsonHierarchyRouteSet.as_router())

router_json_order_by = BazisRouter(tags=['Dynamic_JsonOrderBy'])
router_json_order_by.register(routes.JsonOrderByRouteSet.as_router())

router_json_slice = BazisRouter(tags=['Dynamic_JsonSlice'])
router_json_slice.register(routes.JsonSliceRouteSet.as_router())

router_json_many_to_many = BazisRouter(tags=['Dynamic_ManyToMany'])
router_json_many_to_many.register(routes.DriverDivisionsHiredDatesRouteSet.as_router())

router_related_one_relation_one_field = BazisRouter(tags=['Dynamic_RelatedOneRelation'])
router_related_one_relation_one_field.register(
    routes.RelatedOneRelationOneFieldRouteSet.as_router()
)

router_related_one_relation_multiple_fields = BazisRouter(tags=['Dynamic_RelatedOneRelation'])
router_related_one_relation_multiple_fields.register(
    routes.RelatedOneRelationMultipleFieldsRouteSet.as_router()
)

router_related_one_to_one = BazisRouter(tags=['Dynamic_RelatedOneToOne'])
router_related_one_to_one.register(routes.RelatedOneToOneRouteSet.as_router())

router_related_hierarchy_one_branch = BazisRouter(tags=['Dynamic_RelatedHierarchyOneBranch'])
router_related_hierarchy_one_branch.register(routes.RelatedHierarchyRouteSet.as_router())

router_subaggr = BazisRouter(tags=['Dynamic_SubAggr'])
router_subaggr.register(routes.VehicleSubaggrRouteSet.as_router())

router_context = BazisRouter(tags=['Dynamic_Context'])
router_context.register(routes.VehicleContextRouteSet.as_router())

router_errors_related_table = BazisRouter(tags=['Dynamic_ErrorsRelatedTable'])
router_errors_related_table.register(routes.ErrorsRelatedTableRouteSet.as_router())

router_errors_context = BazisRouter(tags=['Dynamic_ErrorsContext'])
router_errors_context.register(routes.ErrorsContextRouteSet.as_router())

routers_with_prefix = {
    'is_exists_one_relation': router_is_exists_one_relation,
    'is_exists_hierarchy': router_is_exists_hierarchy,
    'json_one_relation': router_json_one_relation,
    'json_hierarchy': router_json_hierarchy,
    'json_order_by': router_json_order_by,
    'json_slice': router_json_slice,
    'json_many_to_many': router_json_many_to_many,
    'related_one_relation_one_field': router_related_one_relation_one_field,
    'related_one_relation_multiple_fields': router_related_one_relation_multiple_fields,
    'related_one_to_one': router_related_one_to_one,
    'related_hierarchy_one_branch': router_related_hierarchy_one_branch,
    'subaggr': router_subaggr,
    'context': router_context,
    'errors_related_table': router_errors_related_table,
    'errors_context': router_errors_context,
}



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.core.exceptions import FieldError
from django.db.models import F, QuerySet
from django.utils.translation import gettext as _

from bazis.core.errors import JsonApiBazisError, JsonApiBazisException
from bazis.core.routes_abstract.initial import InitialRouteBase


class SortingSearching:
    """
    Class responsible for handling sorting and searching operations on querysets.

    Tags: RAG, EXPORT
    """

    terms: list[str]
    route_cls: type[InitialRouteBase]

    def __init__(self, sort: str = None):
        """
        Initializes the SortingSearching instance with sorting terms parsed from a
        comma-separated string.
        """
        self.terms = []
        if sort:
            self.terms = [it.strip() for it in sort.split(',') if it.strip()]

    def apply(self, queryset: QuerySet):
        """
        Applies the sorting terms to the given queryset. Raises a JsonApiBazisException
        if the sorting parameters are invalid.
        """
        if not self.terms:
            return queryset
        try:
            return queryset.order_by(
                *[F(t[1:]).desc(nulls_last=True) if t.startswith('-') else t for t in self.terms]
            )
        except FieldError:
            raise JsonApiBazisException(
                JsonApiBazisError(
                    detail=_('Invalid sorting parameters: `%s`') % self.terms, loc=('path', 'sort')
                ),
            ) from None



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from bazis_test_utils.utils import get_api_client

from tests.utils.assert_sql import assert_sql_query, get_sql_query


@pytest.mark.django_db(transaction=True)
def test_context(sample_app, sample_vehicle_data):
    """Test for checking context."""

    organization = sample_vehicle_data['organization']

    response = get_api_client(sample_app).get('/api/v1/context/entity/driver/')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
    SELECT "entity_driver"."dt_created",
           "entity_driver"."dt_updated",
           "entity_driver"."id",
           "entity_driver"."first_name",
           "entity_driver"."last_name",
           "entity_driver"."contact_phone",
           "entity_driver"."org_owner_id",
           ARRAY
      (SELECT JSONB_BUILD_OBJECT(('id')::text, V0."id", ('_jsonapi_type')::text, 'entity.division') AS "json"
       FROM "entity_division" V0
       WHERE EXISTS
           (SELECT 1 AS "a"
            FROM "entity_driver_divisions" U0
            WHERE (U0."division_id" = (V0."id")
                   AND U0."driver_id" = ("entity_driver"."id"))
            LIMIT 1)) AS "divisions__ids",
           ARRAY
      (SELECT JSONB_BUILD_OBJECT(('id')::text, W0."id", ('dt_start')::text, W0."dt_start", ('dt_finish')::text, W0."dt_finish", ('fact_waste_weight')::text, W0."fact_waste_weight", ('vehicle_id')::text, W0."vehicle_id", ('_vehicle')::text, ARRAY
                                   (SELECT JSONB_BUILD_OBJECT(('gnum')::text, V0."gnum", ('dt_created')::text, V0."dt_created", ('_vehicle_model')::text, ARRAY
                                                                (SELECT JSONB_BUILD_OBJECT(('engine_type')::text, U0."engine_type") AS "json"
                                                                 FROM "entity_vehiclemodel" U0
                                                                 WHERE U0."id" = (V0."vehicle_model_id"))) AS "json"
                                    FROM "entity_vehicle" V0
                                    WHERE V0."id" = (W0."vehicle_id"))) AS "json"
       FROM "entity_carriertask" W0
       WHERE (W0."driver_id" = ("entity_driver"."id")
              AND W0."org_owner_id" = 'id_hex')) AS "_carrier_tasks"
    FROM "entity_driver"
    LIMIT 20;"""

    assert_sql_query(expected_sql_template, sql_query, organization.id)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from django.apps import apps

from fastapi import HTTPException

from pydantic import BaseModel

from bazis.core.routes_abstract.initial import http_get, http_patch
from bazis.core.routes_abstract.jsonapi import (
    JsonApiResponse,
    JsonapiRouteBase,
    api_action_jsonapi_init,
)
from bazis.core.schemas.enums import ApiAction, CrudApiAction
from bazis.core.schemas.fields import SchemaFields

from .route_bases import TimestampsJsonapiRouteBase
from .schemas import CarrierTaskResponse


class CarrierTaskRouteSet(JsonapiRouteBase):
    """Route for CarrierTaskRouteSet"""

    model = apps.get_model('route_injection.CarrierTask')


class VehicleModelRouteBase(JsonapiRouteBase):
    """Route for VehicleModelRouteBase"""

    model = apps.get_model('route_injection.VehicleModel')


class VehicleBrandRouteBase(JsonapiRouteBase):
    """Route for VehicleBrandRouteBase"""

    model = apps.get_model('route_injection.VehicleBrand')

    fields: dict[ApiAction, SchemaFields] = {
        None: SchemaFields(
            exclude={'dt_created': None},
        ),
    }


class VehicleCurrentCarrierTaskRequest(BaseModel):
    driver_id: str | None = None
    dt_finish: datetime | None = None


class VehicleRouteBase(TimestampsJsonapiRouteBase):
    """Route for Vehicle"""

    model = apps.get_model('route_injection.Vehicle')


class VehicleCarrierTaskRouteBase(JsonapiRouteBase):
    """Route for Vehicle"""

    model = apps.get_model('route_injection.Vehicle')

    @http_get('/some_statics/', inject_tags=[CrudApiAction.LIST])
    def get_some_statics(self, **kwargs) -> list[dict]:
        # Check with correct response description
        return [{'k1': 'v1'}, {'k2': 'v2'}]

    @http_get('/some_statics_negative_1/', inject_tags=[CrudApiAction.LIST])
    def get_some_statics_negative_1(self, **kwargs):
        # Reproduce error when response description is missing
        return [{'k1': 'v1'}, {'k2': 'v2'}]

    @http_get('/some_statics_negative_2/', inject_tags=[CrudApiAction.LIST])
    def get_some_statics_negative_2(self, **kwargs) -> dict:
        # Reproduce error when response description is incorrect
        return [{'k1': 'v1'}, {'k2': 'v2'}]

    @http_get(
        '/{item_id}/current_carrier_task/',
        response_model=CarrierTaskResponse,  # pydantic model
        endpoint_callbacks=[api_action_jsonapi_init],
    )
    def get_current_carrier_task(self, item_id: str, **kwargs):
        vehicle = self.set_item(item_id)

        task_queryset = vehicle.get_tasks_for_vehicle(item_id)

        if not task_queryset.exists():
            raise HTTPException(status_code=404, detail='Tasks not found')

        task = task_queryset.first()

        return {
            'data': {
                'id': str(task.id),
                'type': 'route_injection.carrier_task',
                'bs:action': 'view',
                'attributes': {
                    'dt_created': task.dt_created.isoformat(),
                    'dt_updated': task.dt_updated.isoformat(),
                    'dt_start': task.dt_start.isoformat() if task.dt_start else None,
                    'dt_finish': task.dt_finish.isoformat() if task.dt_finish else None,
                    'fact_waste_weight': (
                        str(task.fact_waste_weight) if task.fact_waste_weight else None
                    ),
                },
                'relationships': {
                    'vehicle': {
                        'data': {
                            'id': str(task.vehicle.id),
                            'type': 'route_injection.vehicle',
                        }
                    },
                    'driver': {
                        'data': {
                            'id': str(task.driver.id),
                            'type': 'route_injection.driver',
                        }
                    },
                    'org_owner': {
                        'data': {
                            'id': str(task.org_owner.id),
                            'type': 'route_injection.organization',
                        }
                    },
                },
            }
        }

    @http_patch(
        '/{item_id}/current_carrier_task/',
        response_model_route=CarrierTaskRouteSet.action_retrieve,  # wrapped in an @http_ RouteBase method whose response model is needed
    )
    def patch_current_carrier_task(
        self, item_id: str, data: VehicleCurrentCarrierTaskRequest, **kwargs
    ):
        vehicle = self.set_item(item_id)
        tasks_queryset = vehicle.get_tasks_for_vehicle(item_id)

        if not tasks_queryset.exists():
            raise HTTPException(status_code=404, detail='Tasks not found')

        task = tasks_queryset.first()

        if data.dt_finish:
            task.dt_finish = data.dt_finish
        if data.driver_id:
            task.driver_id = data.driver_id

        task.save()

        return task

    @http_get(
        '/{item_id}/first_carrier_task/',
        response_model=CarrierTaskRouteSet.endpoint_make(
            CarrierTaskRouteSet.routes_ctx['action_retrieve']
        ).route_params.response_model,
        response_model_exclude_unset=True,
        response_class=JsonApiResponse,
    )
    def get_last_carrier_task(self, item_id: str, **kwargs):
        vehicle = self.set_item(item_id)

        task_queryset = vehicle.get_tasks_for_vehicle(item_id)

        if not task_queryset.exists():
            raise HTTPException(status_code=404, detail='Tasks not found')

        task = task_queryset.last()

        return task



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bazis framework core configuration module.

Tags: RAG, INTERNAL
"""

# ruff: noqa: N806

# Monkey patching moved here to avoid circular import issues.
# Placing in bazis/__init__.py causes infinite loop in top-level module linking.

import logging
import os
import sys
from copy import deepcopy
from importlib import import_module
from typing import Any, cast

from django.db import connections
from django.utils.translation import gettext_lazy as _

from pydantic import create_model

from dotenv import load_dotenv

import bazis
import bazis.core.utils.fastapi_monkey_patch
from bazis.core import constance_conf
from bazis.core.utils.imp import get_modules_from_pkg
from bazis.core.utils.locale import discover_locale_paths
from bazis.core.utils.logging_level import force_global_logging_level
from bazis.core.utils.schemas import BazisSettings


try:
    import bazis.contrib
except ImportError:
    pass

logger = logging.getLogger()

# Auto-detect project structure from DJANGO_SETTINGS_MODULE
PROJECT_MODULE = None
SETTINGS_MODULE = None
BASE_DIR = None
if settings_module_path := os.getenv('DJANGO_SETTINGS_MODULE'):
    SETTINGS_MODULE = import_module(settings_module_path)
    if project_module_path := '.'.join(settings_module_path.split('.')[:-1]):
        PROJECT_MODULE = import_module('.'.join(settings_module_path.split('.')[:-1]))
        BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(PROJECT_MODULE.__file__)))
    else:
        BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(SETTINGS_MODULE.__file__)))

# Load environment files with priority preservation
env_files = BazisSettings.model_config['env_file']
if isinstance(env_files, str):
    env_files = [env_files]

# Preserve system environment variables (highest priority)
sys_envs = deepcopy(os.environ)

# Load .env files in order
for env_file in env_files:
    if os.path.exists(env_file):
        load_dotenv(dotenv_path=env_file, override=True)

# Restore system env vars to ensure they override .env files
for k, v in sys_envs.items():
    if v is not None:
        os.environ[k] = v

# Configure logging level from BS_LOG_LEVEL environment variable
log_level = logging.INFO
if os.environ.get('BS_LOG_LEVEL') is not None:
    log_level_name = os.environ['BS_LOG_LEVEL'].upper()
    if log_level_name in logging._nameToLevel:
        log_level = logging._nameToLevel[log_level_name]
    else:
        logger.warning(f'Invalid log level: {log_level_name}. Using default level: WARNING.')

force_global_logging_level(log_level)
logging.basicConfig(
    level=log_level, format='%(asctime)s - %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S'
)

# Set BASE_DIR and expand __BASE_DIR__ placeholders in environment variables
os.environ.setdefault('BS_BASE_DIR', BASE_DIR)
for key, value in os.environ.items():
    if key.startswith('BS_') and '__BASE_DIR__' in value:
        os.environ[key] = value.replace('__BASE_DIR__', BASE_DIR)

# Core Django apps required by Bazis framework
DEFAULT_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.admin',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.postgres',
    'bazis.core',
    # sequences app is necessary for UniqNumberMixin
    'sequences.apps.SequencesConfig',
    'rangefilter',
    'admin_auto_filters',
    'pgtrigger',
    # Constance: dynamic settings via admin panel and database
    'constance',
    'constance.backends.database',
]

# Core Django middleware required by Bazis framework
DEFAULT_MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]


def conf_modules():
    """
    Discovers and yields all configuration modules from Bazis framework and project.

    Search order (reverse priority):
    1. bazis.core.conf
    2. Bazis contrib apps (from BS_BAZIS_CONFIG_APPS or BS_BAZIS_APPS)
    3. Project-level conf modules

    Yields:
        Configuration modules containing Settings classes

    RAG keywords: config discovery, settings modules, conf modules,
                  configuration discovery, settings assembly
    """
    # Core Bazis configuration
    for conf in get_modules_from_pkg(import_module('bazis.core'), 'conf', first_level_only=True):
        yield conf

    # Bazis contrib apps configuration
    BAZIS_CONFIG_APPS = os.environ.get('BS_BAZIS_CONFIG_APPS') or os.environ.get('BS_BAZIS_APPS')
    if BAZIS_CONFIG_APPS:
        BAZIS_CONFIG_APPS = reversed(eval(BAZIS_CONFIG_APPS))

        for bazis_app_name in BAZIS_CONFIG_APPS:
            for conf in get_modules_from_pkg(
                import_module(bazis_app_name), 'conf', first_level_only=True
            ):
                yield conf
    elif hasattr(bazis, 'contrib'):
        for conf in get_modules_from_pkg(bazis.contrib, 'conf'):
            yield conf

    # Project configuration (highest priority)
    if PROJECT_MODULE:
        for conf in get_modules_from_pkg(PROJECT_MODULE, 'conf'):
            yield conf


# Dynamically create unified Settings class from all discovered conf modules
# Uses Pydantic's create_model to merge Settings classes via multiple inheritance
Settings = cast(
    Any,
    create_model(
        'Settings',
        __base__=tuple(conf.Settings for conf in conf_modules() if hasattr(conf, 'Settings')),
    ),
)

# Instantiate local settings object from merged Settings class
_settings = Settings()

# Constance configuration for dynamic settings editable via admin panel
CONSTANCE_CONFIG = {
    # TODO: Temporary setting, should be configurable per installation
    'BAZIS_TIME_ZONE': (
        'Europe/Moscow',
        _('System time zone for generating agent report'),
        'timezone_select',
    )
}

if SETTINGS_MODULE:
    try:
        from django import conf
        from django.conf import LazySettings
    except ImportError as e:
        logger.error(f'Import error: {e}')
    else:
        # Build CONSTANCE_CONFIG from Settings fields marked as dynamic
        # Fields with json_schema_extra={'dynamic': True} become admin-editable
        for field_name, field_info in Settings.model_fields.items():
            if field_info.json_schema_extra and field_info.json_schema_extra.get('dynamic'):
                CONSTANCE_CONFIG[field_name] = (
                    getattr(_settings, field_name),
                    field_info.title,
                    field_info.annotation,
                )

        # Merge Bazis settings into Django SETTINGS_MODULE
        # Strategy: update existing collections, set missing values
        for sett_key, sett_value in _settings.model_dump().items():
            if sett_key not in SETTINGS_MODULE.__dict__:
                SETTINGS_MODULE.__dict__[sett_key] = sett_value
            else:
                existing_value = SETTINGS_MODULE.__dict__[sett_key]

                # Merge collections intelligently
                if isinstance(existing_value, dict) and isinstance(sett_value, dict):
                    existing_value.update(sett_value)
                elif isinstance(existing_value, list) and isinstance(sett_value, list):
                    existing_value.extend(sett_value)
                elif isinstance(existing_value, set) and isinstance(sett_value, set):
                    existing_value.update(sett_value)
                elif isinstance(existing_value, tuple) and isinstance(sett_value, tuple):
                    SETTINGS_MODULE.__dict__[sett_key] = existing_value + sett_value

        SETTINGS_MODULE.__dict__['CONSTANCE_CONFIG'] = CONSTANCE_CONFIG
        SETTINGS_MODULE.__dict__.setdefault(
            'CONSTANCE_BACKEND', 'constance.backends.database.DatabaseBackend'
        )

        # Ensure all required apps are in INSTALLED_APPS
        INSTALLED_APPS = SETTINGS_MODULE.__dict__.setdefault('INSTALLED_APPS', [])
        for _i, default_app in enumerate(DEFAULT_APPS):
            if default_app not in INSTALLED_APPS:
                INSTALLED_APPS.append(default_app)

        # Ensure all required middleware in MIDDLEWARE
        MIDDLEWARE_NEW = SETTINGS_MODULE.__dict__.get('MIDDLEWARE', [])
        MIDDLEWARE = SETTINGS_MODULE.__dict__['MIDDLEWARE'] = []
        for default_middleware in DEFAULT_MIDDLEWARE:
            if default_middleware not in MIDDLEWARE:
                MIDDLEWARE.append(default_middleware)
        for midl in MIDDLEWARE_NEW:
            if midl not in MIDDLEWARE:
                MIDDLEWARE.append(midl)

        # Auto-discover locale paths for i18n
        SETTINGS_MODULE.__dict__['LOCALE_PATHS'] = discover_locale_paths(BASE_DIR)

        django_settings = LazySettings()

        class DjangoSettingsWrapper:
            """
            Three-tier settings accessor: Constance DB → settings.py → .env

            Provides dynamic settings through Constance (admin-editable) while
            maintaining backward compatibility with static Django settings.

            Database check prevents errors during tests when DB not initialized.

            RAG keywords: settings wrapper, constance, dynamic settings,
                          database settings, admin editable settings
            """

            def __getattribute__(self, name):
                """
                Retrieves setting value with priority: Constance DB > Django settings.
                Checks database availability before Constance lookup (for pytest).
                """
                if name in CONSTANCE_CONFIG:
                    try:
                        # Verify DB connection ready (important for tests)
                        for conn in connections.all(initialized_only=True)[:1]:
                            conn.cursor().execute('select 1')
                    except Exception as e:
                        logger.info(
                            f"""DjangoSettingsWrapper raise exception with not ready db connection {e}\n
                                       if you run test - do not worry about this message, otherwise - it is PROBLEM!
                                    """
                        )
                        return ''
                    return getattr(constance_conf.config, name)
                else:
                    return getattr(django_settings, name)

            def __setattr__(self, name, value):
                """
                Sets setting value in Constance DB if dynamic, else Django settings.
                """
                if name in CONSTANCE_CONFIG:
                    setattr(constance_conf.config, name, value)
                else:
                    setattr(django_settings, name, value)

            def __delattr__(self, name):
                try:
                    object.__delattr__(self, name)
                except AttributeError:
                    pass

            def __iter__(self):
                """
                Iterates all settings from both Constance and Django settings.
                """
                for key in CONSTANCE_CONFIG:
                    yield (key, getattr(constance_conf.config, key))
                for key in django_settings.__dict__.keys():
                    yield (key, getattr(django_settings, key))

        django_settings_wrapper = DjangoSettingsWrapper()

        class ConfWrapper:
            """
            Monkey patches django.conf module to inject DjangoSettingsWrapper.
            Replaces django.conf.settings with our custom wrapper.

            RAG keywords: django conf wrapper, monkey patch django, settings override
            """

            def __getattribute__(self, name):
                """
                Returns DjangoSettingsWrapper for 'settings' attribute.
                Delegates all other attributes to original django.conf module.
                """
                if name == 'settings':
                    return django_settings_wrapper
                return getattr(conf, name)

        # Replace django.conf module with wrapper
        sys.modules['django.conf'] = ConfWrapper()


class SettingsWrapper:
    """
    Unified settings accessor for Bazis framework.
    Prioritizes Constance dynamic settings over static settings.

    Used as global settings object throughout application.

    RAG keywords: settings accessor, bazis settings, settings wrapper
    """

    def __getattribute__(self, name):
        """
        Retrieves setting: Constance DB if dynamic, else static Settings.
        """
        if name in CONSTANCE_CONFIG:
            return getattr(constance_conf.config, name)
        else:
            return getattr(_settings, name)


# Global settings instance for application use
settings = SettingsWrapper()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING

from fastapi.responses import JSONResponse

from starlette.status import HTTP_400_BAD_REQUEST, HTTP_422_UNPROCESSABLE_ENTITY

from bazis.core.errors import SchemaErrors
from bazis.core.routes_abstract.context import RouteContext
from bazis.core.schemas.enums import ApiAction, CrudApiAction
from bazis.core.schemas.fields import SchemaMetaFields
from bazis.core.schemas.meta import get_meta_schemas
from bazis.core.utils.django_types import TYPES_DJANGO_TO_SCHEMA_LOOKUP
from bazis.core.utils.functools import func_sig_param_replace


if TYPE_CHECKING:
    from .route_base import JsonapiRouteBase


def meta_fields_addition(
    cls: type['JsonapiRouteBase'], route_ctx: RouteContext, api_action: ApiAction
):
    """
    Function for adding meta-fields to a JSON:API route based on context classes.
        Extracts meta-fields from the route's context classes and adds them to
        the route class's meta_fields for the specified API action.
        Called as a callback when initializing the route for an API action.

        Example of usage in a route method:
            ```python
            @http_get('/{item_id}/', inject_tags=[CrudApiAction.RETRIEVE], endpoint_callbacks=[
                partial(meta_fields_addition, api_action=CrudApiAction.RETRIEVE),
                partial(api_action_init, api_action=CrudApiAction.RETRIEVE),
                partial(api_action_response_init, api_action=CrudApiAction.RETRIEVE),
                api_action_jsonapi_init,
                item_id_typing,
            ])
            def action_retrieve(self, item_id: str, **kwargs):
                return {
                    'data': self.retrieve(str(item_id).strip()),
                    'meta': self.get_meta_fields(CrudApiAction.RETRIEVE),
                    'included': chain(*self.item.fields_for_included.values()),
                }
            ```

        :param cls: JsonapiRouteBase class
        :param route_ctx: Route context object
        :param api_action: HTTP action within which we extract meta-fields

    Tags: RAG, EXPORT
    """
    # collect information about meta-fields from context classes
    for ctx_cls in cls.get_context_classes(route_ctx).values():
        for name, meta_schema in get_meta_schemas(ctx_cls).items():
            if api_action in meta_schema.api_actions:
                if api_action not in cls.meta_fields:
                    cls.meta_fields[api_action] = SchemaMetaFields()
                cls.meta_fields[api_action].include[name] = meta_schema.field_schema


def build_schema_defaults(cls, api_action: ApiAction):
    """
    Creates and stores default schemas for the specified API action.
        If a schema for the API action already exists, it is not recreated.
        For the UPDATE action, the schema is created by combining the inclusions
        of the schemas for the UPDATE and CREATE actions.
    Schemas are created using the schema factory associated with the route class.
    Used similarly to meta_fields_addition.

    Tags: RAG
    """
    if api_action == CrudApiAction.UPDATE:
        if CrudApiAction.UPDATE not in cls.schema_factories:
            cls.schema_factories[CrudApiAction.UPDATE] = cls.build_schema_factory(
                CrudApiAction.UPDATE
            )
        if CrudApiAction.CREATE not in cls.schema_factories:
            cls.schema_factories[CrudApiAction.CREATE] = cls.build_schema_factory(
                CrudApiAction.CREATE
            )

        if CrudApiAction.UPDATE not in cls.schema_defaults:
            cls.schema_defaults[CrudApiAction.UPDATE] = cls.schema_factories[
                CrudApiAction.UPDATE
            ].build_schema(
                inclusions=(
                    cls.schema_factories[CrudApiAction.UPDATE].inclusions_list
                    + cls.schema_factories[CrudApiAction.CREATE].inclusions_list
                )
            )
    else:
        if api_action not in cls.schema_factories:
            cls.schema_factories[api_action] = cls.build_schema_factory(api_action)
        if api_action not in cls.schema_defaults:
            cls.schema_defaults[api_action] = cls.schema_factories[api_action].build_schema()
            cls.schema_responses_defaults[api_action] = cls.schema_factories[
                api_action
            ].build_schema(is_response_schema=True)


def api_action_init(cls, route_ctx: RouteContext, api_action: ApiAction):
    """
    Triggers building of the default schema for the specified API action.
    Stores the API action in the route context.
    Used similarly to meta_fields_addition.

    Tags: RAG, EXPORT
    """
    build_schema_defaults(cls, api_action)
    route_ctx.store['api_action'] = api_action


def api_action_response_init(cls, route_ctx: RouteContext, api_action: ApiAction):
    """
    Sets the response model for the route based on the default schema.
    Used similarly to meta_fields_addition.

    Tags: RAG, EXPORT
    """
    build_schema_defaults(cls, api_action)
    route_ctx.route_params.response_model = cls.schema_defaults[api_action]
    route_ctx.store['api_action_response'] = api_action


class JsonApiResponse(JSONResponse):
    """
    JSON:API response type with media_type 'application/vnd.api+json'.

    Tags: RAG, EXPORT
    """

    media_type = 'application/vnd.api+json'


def api_action_jsonapi_init(cls, route_ctx: RouteContext):
    """
    Converts the route into a JSON:API route by setting the appropriate response class
    and exception parameters for the response model.
    Used similarly to meta_fields_addition.

    Tags: RAG, EXPORT
    """
    # without this setting, fields removed from the private schema with a value of null
    # are included in the output attributes
    route_ctx.route_params.response_model_exclude_unset = True
    route_ctx.route_params.response_class = JsonApiResponse
    route_ctx.route_params.responses = {
        HTTP_400_BAD_REQUEST: {'model': SchemaErrors},
        HTTP_422_UNPROCESSABLE_ENTITY: {'model': SchemaErrors},
    }


def item_data_typing(cls, route_ctx: RouteContext, api_action: ApiAction):
    """
    Replaces the item_data parameter annotation with the default schema for the API action.
    Used similarly to meta_fields_addition.

    Tags: RAG, EXPORT
    """
    func_sig_param_replace(
        route_ctx.endpoint, 'item_data', annotation=cls.schema_defaults[api_action]
    )


def item_id_typing(cls, route_ctx: RouteContext):
    """
    Replaces the item_id parameter annotation with the primary key type from the read schema.
    Used similarly to meta_fields_addition.

    Tags: RAG, EXPORT
    """
    func_sig_param_replace(
        route_ctx.endpoint,
        'item_id',
        annotation=TYPES_DJANGO_TO_SCHEMA_LOOKUP[cls.model._meta.pk],
    )


def crud_jsonapi_callbacks(api_action: CrudApiAction) -> list[Callable]:
    """
    Returns a list of endpoint callbacks for the given CrudApiAction.

    Replaces the standard JSONResponse with JsonApiResponse, which has media_type = 'application/vnd.api+json'.
    Adds descriptions of 400 and 422 errors to the OpenAPI documentation (SchemaErrors).
    Removes fields with the value None from the response — in JSON:API they must not be present.

    Tags: RAG
    """
    callbacks = [api_action_jsonapi_init]

    if api_action != CrudApiAction.DESTROY:
        callbacks.extend(
            [
                # adding meta, in most cases pagination for a list, but it can be an addition to any schema
                partial(meta_fields_addition, api_action=api_action),
                # building a schema for api_action
                partial(api_action_init, api_action=api_action),
                # setting the response model
                partial(
                    api_action_response_init,
                    api_action=(
                        api_action if api_action is CrudApiAction.LIST else CrudApiAction.RETRIEVE
                    ),
                ),
            ]
        )

    if api_action in {CrudApiAction.CREATE, CrudApiAction.UPDATE}:
        # create and update have a request body
        # therefore it is necessary to substitute a specific Pydantic schema of the request body into the route signature,
        # so that everything is validated correctly and displayed in Swagger
        callbacks.append(partial(item_data_typing, api_action=api_action))

    if api_action in {CrudApiAction.RETRIEVE, CrudApiAction.UPDATE, CrudApiAction.DESTROY}:
        #  clarification of the item_id parameter type in the route signature based on the RETRIEVE schema
        #  substitutes the type of the primary key field (usually str, int, UUID, etc.),
        #  so that FastAPI correctly validates item_id when it is passed in the path
        callbacks.append(item_id_typing)

    return callbacks



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from bazis_test_utils.utils import get_api_client


@pytest.mark.django_db(transaction=True)
def test_context_errors(sample_app, dynamic_vehicle_data):
    """Test to check for the absence of the declared context."""

    with pytest.raises(KeyError, match="Missing keys in the request context: {'_user'}"):
        get_api_client(sample_app).get('/api/v1/errors_context/dynamic/division/')

        pytest.fail('A KeyError was expected, but the request completed successfully.')


@pytest.mark.django_db(transaction=True)
def test_field_warning_logged(sample_app, dynamic_vehicle_data, caplog):
    """Check that a warning about the missing relation is written to the log."""

    with caplog.at_level('WARNING'):
        response = get_api_client(sample_app).get('/api/v1/errors_related_table/dynamic/division/')

    assert response.status_code == 200
    assert any(
        'Error in calculated field drivers_list1: model Division does not have related table drivers1'
        in message
        for message in caplog.messages
    )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
from decimal import Decimal

import django
from django.utils.timezone import now

from route_injection.models import (
    CarrierTask,
    Country,
    Driver,
    Organization,
    Vehicle,
    VehicleBrand,
    VehicleModel,
)


os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'sample.settings')
django.setup()

NUM_ORGS = 5
NUM_COUNTRIES = 5
NUM_BRANDS = 10
NUM_MODELS = 10
NUM_VEHICLES = 100
NUM_DRIVERS = 300
TASKS_PER_DRIVER = 10


def generate_test_data():
    print('Creating organizations, brands, models, countries...')

    orgs = [Organization.objects.create(name=f'Org __#{i}') for i in range(NUM_ORGS)]

    [Country.objects.create(name=f'Country __#{i}') for i in range(NUM_COUNTRIES)]

    brands = [VehicleBrand.objects.create(name=f'Brand __#{i}') for i in range(NUM_BRANDS)]

    models = [
        VehicleModel.objects.create(
            brand=random.choice(brands),
            model=f'Model __#{i}',
            engine_type=random.choice(['Diesel', 'Petrol', 'Electric']),
            capacity=Decimal(random.uniform(1.0, 10.0)).quantize(Decimal('0.01')),
        )
        for i in range(NUM_MODELS)
    ]

    print('Creating vehicles...')

    vehicles = [
        Vehicle.objects.create(vehicle_model=random.choice(models), gnum=f'VEH-{i:05d}')
        for i in range(NUM_VEHICLES)
    ]

    print('Creating drivers and trips...')

    for i in range(NUM_DRIVERS):
        driver = Driver.objects.create(
            first_name=f'Driver{i}',
            last_name='Testov',
            contact_phone=f'8-977-899-{1000 + i}',
            org_owner=random.choice(orgs),
        )

        vehicle = random.choice(vehicles)

        for j in range(TASKS_PER_DRIVER):
            CarrierTask.objects.create(
                vehicle=vehicle,
                driver=driver,
                dt_start=now(),
                dt_finish=None if j % 2 == 0 else now(),
                fact_waste_weight=Decimal(random.uniform(1.0, 10.0)).quantize(Decimal('0.1')),
                org_owner=driver.org_owner,
            )

        if i % 200 == 0:
            print(f'Drivers created: {i}')

    print(
        f'Total: {NUM_DRIVERS} drivers, {NUM_DRIVERS * TASKS_PER_DRIVER} trips, {NUM_VEHICLES} vehicles.'
    )


if __name__ == '__main__':
    generate_test_data()



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Application factory and singleton holder for the Bazis project.

Tags: RAG, EXPORT
"""

import importlib
import logging
import sys
import types


_STATE_KEY = 'bazis.core._app_singleton'
_state = sys.modules.get(_STATE_KEY)
if _state is None:
    _state = types.SimpleNamespace(app=None, initializing=False, initialized=False)
    sys.modules[_STATE_KEY] = _state


def get_app_base():
    if _state.app is None:
        _state.app = _create_app_base()
    return _state.app


def ensure_app_initialized():
    app = get_app_base()
    if not _state.initialized and not _state.initializing:
        _state.initializing = True
        try:
            _initialize_app(app)
            _state.initialized = True
        finally:
            _state.initializing = False
    return app


def get_app():
    return ensure_app_initialized()


def _create_app_base():
    # ruff: noqa: E402
    import os
    import sys as _sys

    import django

    _sys.path.append(os.getcwd())
    django.setup()

    from django.conf import settings

    from fastapi import FastAPI

    LOG = logging.getLogger() # noqa: N806

    if BAZIS_APP_MODULE := getattr(settings, 'BAZIS_APP_MODULE', None): # noqa: N806
        app_module = importlib.import_module(BAZIS_APP_MODULE)
        app = app_module.app
    else:
        LOG.info('Custom application not found. Default will be created')
        if settings.DEBUG:
            app = FastAPI(
                openapi_url='/api/openapi.json',
                docs_url='/api/swagger/',
                redoc_url='/api/redoc/',
                swagger_ui_oauth2_redirect_url='/api/swagger/oauth2-redirect',
                swagger_ui_parameters={'defaultModelsExpandDepth': 0},
            )
        else:
            app = FastAPI(
                openapi_url=None,
                docs_url=None,
                redoc_url=None,
                swagger_ui_oauth2_redirect_url=None,
                swagger_ui_parameters={'defaultModelsExpandDepth': 0},
            )

    return app


def _initialize_app(app): # noqa: C901
    # ruff: noqa: E402
    import os
    import traceback

    from django.conf import settings
    from django.utils.translation import get_language, to_locale

    from fastapi import Request
    from fastapi.encoders import jsonable_encoder
    from fastapi.exceptions import HTTPException, RequestValidationError
    from fastapi.middleware.cors import CORSMiddleware
    from fastapi.responses import RedirectResponse, Response

    from starlette.concurrency import run_in_threadpool
    from starlette.middleware.sessions import SessionMiddleware
    from starlette.responses import JSONResponse
    from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

    from bazis.core.i18n import LanguageMiddleware, expand_lang
    from bazis.core.utils.functools import get_attr
    from bazis.core.utils.orm import close_old_connections

    from .errors import JsonApiBazisException, SchemaError, SchemaErrors, SchemaErrorSource

    @app.get(f'{settings.MEDIA_URL}{{path:path}}')
    async def redirect_media(path: str):
        if settings.MEDIA_HOST_URL:
            return RedirectResponse(url=f'{settings.MEDIA_HOST_URL}{settings.MEDIA_URL}{path}')
        return RedirectResponse(url=f'{settings.ADMIN_HOST_URL}{settings.MEDIA_URL}{path}')

    @app.get(f'{settings.STATIC_URL}{{path:path}}')
    async def redirect_static(path: str):
        return RedirectResponse(url=f'{settings.ADMIN_HOST_URL}{settings.STATIC_URL}{path}')

    class CloseOldConnectionsMiddleware:
        """
        Middleware that checks and closes dropped connections to maintain database
        connection integrity.
        """

        def __init__(self, app) -> None:
            """
            Initializes the CloseOldConnectionsMiddleware with the given application
            instance.
            """
            self.app = app

        async def __call__(self, scope, receive, send) -> None:
            """
            Executes the middleware, ensuring old connections are closed before and after
            handling the request.
            """
            await run_in_threadpool(close_old_connections)
            await self.app(scope, receive, send)
            await run_in_threadpool(close_old_connections)

    app.add_middleware(CloseOldConnectionsMiddleware)

    app.add_middleware(LanguageMiddleware)

    app.add_middleware(SessionMiddleware, secret_key=settings.SECRET_KEY)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.CSRF_TRUSTED_ORIGINS,
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )

    @app.router.get('/api/schemas.json')
    async def get_api_schemas(request: Request):
        """
        Fetches the API schemas for the current language and redirects to the
        corresponding JSON schema file.
        """
        for lang in expand_lang(to_locale(get_language())):
            lang_path = os.path.join(settings.STATIC_ROOT, f'schemas_{lang}.json')
            if os.path.exists(lang_path):
                lang_path_dt = os.path.getmtime(lang_path)
                return RedirectResponse(
                    f'{settings.STATIC_URL}schemas_{lang}.json?_h={lang_path_dt}'
                )

    @app.router.get('/api/healthcheck')
    async def healthcheck(request: Request):
        """
        Returns an empty response to indicate that the application is healthy.
        """
        return Response('')

    def get_source_from_loc(
        loc: list | tuple | None, _id: str = None, _type: str = None
    ) -> SchemaErrorSource | None:
        """
        Generates a SchemaErrorSource object from the given location, ID, and type
        information.
        """
        attrs = {}

        if _id:
            attrs['id'] = str(_id)
        if _type:
            attrs['type'] = _type

        if loc:
            loc = [str(x) for x in loc]
            if loc[0] == 'path':
                attrs['parameter'] = '/'.join([''] + loc[1:])
            else:
                if loc[0] == 'body':
                    loc = loc[1:]
                attrs['pointer'] = '/'.join([''] + loc)

        if not attrs:
            return None

        return SchemaErrorSource(**attrs)

    def exc_encoder(
        errs: list[SchemaError], status: int, cookies: list[tuple[str, str, int]] = None
    ):
        """
        Encodes a list of SchemaError objects into a JSON response with the specified
        status and optional cookies.
        """
        response = JSONResponse(
            jsonable_encoder(SchemaErrors(errors=errs), exclude_unset=True, exclude_none=True),
            status_code=status,
        )

        if cookies:
            for cookie_name, cookie_value, cookie_age in cookies:
                response.set_cookie(key=cookie_name, value=cookie_value, max_age=cookie_age)

        return response

    @app.exception_handler(JsonApiBazisException)
    async def json_api_bazis_exception_handler(
        request: Request, exc: JsonApiBazisException
    ) -> JSONResponse:
        """
        Handles JsonApiBazisException by converting it to a JSONAPI-compliant JSONResponse.
        :param request: The current request object.
        :param exc: The exception object.
        :return: JSONResponse.
        """

        def get_item_id(err):
            """
            Retrieves the ID of the item associated with the error, if available.
            """
            if not err.item:
                return None
            return err.item.id

        def get_item_type(err):
            """
            Retrieves the resource label of the item associated with the error, if
            available.
            """
            if not err.item:
                return None
            return err.item.get_resource_label()

        return exc_encoder(
            [
                SchemaError(
                    status=err.status,
                    code=err.code,
                    title=str(err.title) if err.title else None,
                    detail=str(err.detail) if err.detail else None,
                    source=get_source_from_loc(
                        err.loc, _id=get_item_id(err), _type=get_item_type(err)
                    ),
                    meta=err.meta,
                )
                for err in exc.errors
            ],
            exc.status,
            exc.cookies,
        )

    @app.exception_handler(RequestValidationError)
    async def json_api_request_validation_exception_handler(
        request: Request, exc: RequestValidationError
    ) -> JSONResponse:
        """
        Handles RequestValidationError by converting it to a JSONAPI-compliant JSONResponse.
        :param request: The current request object.
        :param exc: The exception object.
        :return: JSONResponse.
        """
        return exc_encoder(
            [
                SchemaError(
                    status=HTTP_422_UNPROCESSABLE_ENTITY,
                    code='ERR_VALIDATE',
                    title=err['type'],
                    detail=err['msg'],
                    source=get_source_from_loc(
                        err.get('loc'),
                        _id=get_attr(err, 'ctx._id'),
                        _type=get_attr(err, 'ctx._type'),
                    ),
                )
                for err in exc.errors()
            ],
            HTTP_422_UNPROCESSABLE_ENTITY,
        )

    @app.exception_handler(HTTPException)
    async def json_api_http_exception_handler(
        request: Request, exc: HTTPException
    ) -> JSONResponse:
        """
        Handles common HTTP exceptions by converting them to a JSONAPI-compliant JSONResponse.
        :param request: The current request object.
        :param exc: The exception object.
        :return: JSONResponse.
        """
        meta = {}

        headers = getattr(exc, 'headers', None)
        if headers:
            meta.update({'headers': getattr(exc, 'headers', None)})

        meta = meta or None

        return exc_encoder(
            [
                SchemaError(
                    status=exc.status_code,
                    detail=str(exc.detail) if exc.detail else None,
                    meta=meta,
                    traceback=''.join(traceback.format_exception(exc))
                    if settings.DEBUG
                    else None,
                )
            ],
            exc.status_code,
        )

    @app.exception_handler(500)
    async def json_api_http_500_handler(
        request: Request, exc: Exception
    ) -> JSONResponse:  # noqa: F811
        """
        Handles generic exceptions by converting them to a JSONAPI-compliant JSONResponse.
        :param request: The current request object.
        :param exc: The exception object.
        :return: JSONResponse.
        """
        return exc_encoder(
            [
                SchemaError(
                    status=500,
                    detail=traceback.format_exc() if settings.DEBUG else None,
                )
            ],
            500,
        )

    from bazis.core.router import router
    from bazis.core.routing import BazisRoute

    router.routes_cast(BazisRoute)
    app.include_router(router)



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Schema definitions for route filter fields.

This module defines:
- `InputField`: a representation of a single filterable field with name and OpenAPI type.
- `FieldListModel`: a container for returning multiple `InputField` instances.
- `RouteFilterFieldsSchemas`: a custom OpenAPI schema generator that overrides the default type mapping
  for specific types such as `Decimal`, `date`, and `datetime`.

Used in conjunction with `RouteFilterFieldsService` to expose filter metadata via API.
"""


from pydantic import BaseModel
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue

from pydantic_core import core_schema


class InputField(BaseModel):
    """Represents a filterable field with its name and inferred OpenAPI type."""

    name: str
    py_type: str


class FieldListModel(BaseModel):
    """Response model for a list of filterable fields."""

    fields: list[InputField]


class RouteFilterFieldsSchemas(GenerateJsonSchema):
    """
    Custom schema generator for route filter fields.

    Overrides how specific Python types are represented in OpenAPI schemas,
    e.g. forces Decimal to appear as 'Decimal' instead of the default 'number | string'.
    """

    def decimal_schema(self, schema: core_schema.DecimalSchema) -> JsonSchemaValue:
        """
        Override for Decimal fields — represented as custom string 'Decimal'.
        Prevents Pydantic from emitting anyOf[number, string].
        """
        return {'type': 'Decimal'}

    def date_schema(self, schema: core_schema.DateSchema) -> JsonSchemaValue:
        """
        Override for date fields — represented as 'date'.
        """
        return {'type': 'date'}

    def datetime_schema(self, schema: core_schema.DateSchema) -> JsonSchemaValue:
        """
        Override for datetime fields — represented as 'datetime'.
        """
        return {'type': 'datetime'}


class ResourceIdentifier(BaseModel):
    type: str
    id: str


class RelationshipData(BaseModel):
    """
    Universal model for JSON:API relationships.
    data can be:
    - null → to-one remove
    - object → to-one
    - list of objects → to-many
    """

    data: None | ResourceIdentifier | list[ResourceIdentifier]


openapi_relationship_examples = {
    'to_one': {
        'summary': 'To-one relationship',
        'description': 'Changing a single relationship',
        'value': {'data': {'type': 'string', 'id': 'string'}},
    },
    'to_many': {
        'summary': 'To-many relationship',
        'description': 'Changing multiple relationships',
        'value': {'data': [{'type': 'string', 'id': 'string'}, {'type': 'string', 'id': 'string'}]},
    },
}



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by Django 5.0.14 on 2025-05-22 11:43

import django.db.models.deletion
import model_clone.mixin
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Country',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255, unique=True, verbose_name='Country Name')),
            ],
            options={
                'verbose_name': 'Country',
                'verbose_name_plural': 'Countries',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Division',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Vehicle',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('gnum', models.CharField(max_length=50, unique=True, verbose_name='State Registration Number')),
            ],
            options={
                'verbose_name': 'Vehicle',
                'verbose_name_plural': 'Vehicles',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VehicleBrand',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255, unique=True, verbose_name='Brand Name')),
            ],
            options={
                'verbose_name': 'Vehicle Brand',
                'verbose_name_plural': 'Vehicle Brands',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Driver',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('first_name', models.CharField(max_length=255, verbose_name='First Name')),
                ('last_name', models.CharField(max_length=255, verbose_name='Last Name')),
                ('contact_phone', models.CharField(blank=True, max_length=50, null=True, verbose_name='Phone')),
                ('divisions', models.ManyToManyField(blank=True, related_name='drivers', to='route_injection.division')),
                ('org_owner', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='route_injection.organization')),
            ],
            options={
                'verbose_name': 'Driver',
                'verbose_name_plural': 'Drivers',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.AddField(
            model_name='division',
            name='org_owner',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='route_injection.organization'),
        ),
        migrations.CreateModel(
            name='CarrierTask',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('dt_start', models.DateTimeField(blank=True, null=True, verbose_name='Start Time')),
                ('dt_finish', models.DateTimeField(blank=True, null=True, verbose_name='Start Time')),
                ('fact_waste_weight', models.DecimalField(blank=True, decimal_places=3, max_digits=15, null=True, verbose_name='Fact Waste Weight, t')),
                ('driver', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='carrier_tasks', to='route_injection.driver', verbose_name='Driver')),
                ('org_owner', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='route_injection.organization')),
                ('vehicle', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='carrier_tasks', to='route_injection.vehicle', verbose_name='Vehicle')),
            ],
            options={
                'verbose_name': 'Carrier Task',
                'verbose_name_plural': 'Carrier Tasks',
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VehicleAssembleInfo',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('assembly_plant', models.CharField(max_length=50, verbose_name='Assembly Plant')),
                ('assembly_date', models.DateField(verbose_name='Assembly Date')),
                ('vin', models.CharField(max_length=17, unique=True, verbose_name='VIN')),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='route_injection.country', verbose_name='Country')),
                ('vehicle', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='assemble_info', to='route_injection.vehicle', verbose_name='Vehicle')),
            ],
            options={
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VehicleModel',
            fields=[
                ('dt_created', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation time')),
                ('dt_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Update time')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('model', models.CharField(max_length=255, unique=True, verbose_name='Model')),
                ('engine_type', models.CharField(blank=True, max_length=50, null=True, verbose_name='Engine Type')),
                ('capacity', models.DecimalField(blank=True, decimal_places=2, max_digits=6, null=True, verbose_name='Capacity, t')),
                ('brand', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='route_injection.vehiclebrand', verbose_name='Brand')),
            ],
            options={
                'verbose_name': 'Vehicle Model',
                'verbose_name_plural': 'Vehicle Models',
                'abstract': False,
                'unique_together': {('brand', 'model')},
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.AddField(
            model_name='vehicle',
            name='vehicle_model',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='route_injection.vehiclemodel', verbose_name='Vehicle Model'),
        ),
    ]



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.conf import settings
from django.db.models import QuerySet
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from fastapi import Query, Request

from pydantic import BaseModel

from bazis.core.schemas.enums import CrudApiAction
from bazis.core.schemas.meta import meta_field
from bazis.core.schemas.schemas import PaginationLinks


class PaginationMeta(BaseModel):
    """
    Represents the metadata for pagination, including count, limit, and offset.

    Tags: RAG, EXPORT
    """

    count: int
    limit: int
    offset: int


class ServicePagination:
    """
    Handles pagination logic for a given request, including calculating offsets,
    limits, and generating pagination links.

    Tags: RAG, EXPORT
    """

    param_offset: str = 'page[offset]'
    param_limit: str = 'page[limit]'

    def __init__(
        self,
        request: Request,
        offset: int = Query(None, alias=param_offset, ge=0),
        limit: int = Query(
            None, alias=param_limit, ge=0, le=settings.BAZIS_API_PAGINATION_PAGE_SIZE_MAX or 1000
        ),
    ):
        """
        Initializes the ServicePagination instance with request data, offset, limit, and
        default settings.
        """
        self.request = request
        self.offset = offset or 0
        self.limit = settings.BAZIS_API_PAGINATION_PAGE_SIZE_DEFAULT if limit is None else limit
        self.count = 0

    def apply(self, queryset: QuerySet):
        """
        Applies pagination to the given queryset, adjusting it based on the offset and
        limit parameters.
        """
        queryset_count = queryset.prefetch_related(None).select_related(None)
        self.count = queryset_count.count()
        if self.count == 0 or self.offset > self.count:
            return queryset.none()
        if self.limit == 0:
            return queryset.none()
        return queryset[self.offset : self.offset + self.limit]

    @cached_property
    def link_prev(self):
        """
        Generates the URL for the previous page, if applicable, based on the current
        offset and limit.
        """
        if self.limit == 0:
            return None

        if self.offset == 0:
            return None

        if (offset := self.offset - self.limit) <= 0:
            return str(self.request.url.remove_query_params(self.param_offset))

        return str(
            self.request.url.include_query_params(
                **{
                    self.param_limit: self.limit,
                    self.param_offset: offset,
                }
            )
        )

    @cached_property
    def link_next(self):
        """
        Generates the URL for the next page, if applicable, based on the current offset
        and limit.
        """
        if self.limit == 0:
            return None

        if (offset := self.offset + self.limit) >= self.count:
            return None

        return str(
            self.request.url.include_query_params(
                **{
                    self.param_limit: self.limit,
                    self.param_offset: offset,
                }
            )
        )

    @cached_property
    def link_first(self):
        """
        Generates the URL for the first page, removing the offset parameter from the
        query string.
        """
        if self.limit == 0:
            return None
        return str(self.request.url.remove_query_params(self.param_offset))

    @cached_property
    def link_last(self):
        """
        Generates the URL for the last page, calculating the appropriate offset based on
        the total count and limit.
        """
        if self.limit == 0:
            return None
        if self.count % self.limit == 0:
            offset = self.count - self.limit
        else:
            offset = self.count // self.limit * self.limit

        return str(
            self.request.url.include_query_params(
                **{
                    self.param_limit: self.limit,
                    self.param_offset: offset,
                }
            )
        )

    @cached_property
    def links(self):
        """
        Aggregates all pagination links (first, last, previous, next) into a
        PaginationLinks object.
        """
        return PaginationLinks(
            first=self.link_first,
            last=self.link_last,
            prev=self.link_prev,
            next=self.link_next,
        )

    @cached_property
    @meta_field([CrudApiAction.LIST], title=_('Pagination'))
    def pagination(self) -> PaginationMeta:
        """
        Provides pagination metadata, including count, limit, and offset, as a
        PaginationMeta object.
        """
        return PaginationMeta(
            count=self.count,
            limit=self.limit,
            offset=self.offset,
        )



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import unquote_plus

from django.db import models
from django.db.models import QuerySet

from typing_extensions import deprecated

from bazis.core.utils.query_complex import QueryToOrm


class ServiceFiltering:
    """
    Class for applying service-level filtering to a Django QuerySet based on a URL-
    encoded filter string.

    Tags: RAG, EXPORT
    """

    def __init__(self, filter: str = ''):
        """
        Initializes the ServiceFiltering instance with a URL-encoded filter string.

        :param filter: URL-encoded filter string.
        """
        self.query_str = unquote_plus(filter)

    def apply(
        self, queryset: QuerySet, filters_aliases: dict[str, str] = None, fiter_context: dict = None
    ):
        """
        Applies the filtering to the given QuerySet using the specified filter aliases and context.

        :param queryset: The Django QuerySet to filter.
        :param filters_aliases: Optional dictionary mapping filter aliases to actual field names.
        :param fiter_context: Optional context dictionary for additional filtering logic.
        :return: Filtered QuerySet.
        """
        return QueryToOrm.qs_apply(
            queryset,
            self.query_str,
            filters_aliases=filters_aliases,
            fiter_context=fiter_context,
        )


@deprecated('Filtering is deprecated, use bazis.core.utils.query_complex.QueryToOrm instead.')
class Filtering:
    """
    Implementation of filtering for the specified queryset. Filtering is oriented towards
    the composition of the model fields from the queryset.
        Fields can be nested as much as needed, according to Django notation.

        Supported range filters:
        - gt, gte, lt, lte

        Support for special notation for full-text search:

        - '?filter=$search=text': Identical to the query '?search=text'. Full-text search across
        all text and integer fields of the model.
        - '?filter=author__$search=text': Full-text search across all text and integer fields of the
        nested 'author' model.
        - '?filter=author__username__$search=text': Full-text search only on the 'username' text field
        of the nested 'author' entity.

        For filtering non-null fields, the following suffixes are used:

        - For simple fields, the suffix 'isnull': '?filter=name__isnull=true/false'.
        - For relation fields, the suffix 'exists': '?filter=status__exists=true/false'.

    Tags: RAG, EXPORT
    """

    @classmethod
    def qs_apply(
        cls,
        queryset: QuerySet,
        query_str: str,
        filters_aliases: dict[str, str] = None,
        fiter_context: dict = None,
    ):
        return QueryToOrm.qs_apply(
            queryset,
            query_str,
            filters_aliases=filters_aliases,
            fiter_context=fiter_context,
        )

    def __init__(
        self,
        model: models.Model,
        query_str: str,
        filters_aliases: dict[str, str] = None,
        fiter_context: dict = None,
    ):
        self.model = model
        self.query_str = query_str
        self.filters_aliases = filters_aliases or {}
        self.fiter_context = fiter_context or {}

    def __call__(self):
        q_orm = QueryToOrm(self.query_str, self.model, self.filters_aliases, self.fiter_context)
        return q_orm.q, q_orm.fields_calc



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field


class RelationshipData(BaseModel):
    id: str
    type: str


class RelationshipItem(BaseModel):
    data: RelationshipData


class CarrierTaskRelationships(BaseModel):
    vehicle: RelationshipItem
    driver: RelationshipItem
    org_owner: RelationshipItem


class CarrierTaskAttributes(BaseModel):
    dt_created: datetime
    dt_updated: datetime
    dt_start: datetime
    dt_finish: datetime | None
    fact_waste_weight: str | None


class CarrierTaskResponseData(BaseModel):
    id: str
    type: Literal['route_injection.carrier_task']
    bs_action: Literal['view'] = Field(..., alias='bs:action')
    attributes: CarrierTaskAttributes
    relationships: CarrierTaskRelationships

    class Config:
        populate_by_name = True


class CarrierTaskResponse(BaseModel):
    data: CarrierTaskResponseData



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from django.apps import apps

from bazis.core.routes_abstract.jsonapi import JsonapiRouteBase
from bazis.core.schemas.fields import SchemaField, SchemaFields


class PeopleRouteSet(JsonapiRouteBase):
    model = apps.get_model('sparse_fieldsets.People')

    fields = {
        None: SchemaFields(
            include={
                'articles': None,
            },
        ),
    }


class CategoryRouteSet(JsonapiRouteBase):
    model = apps.get_model('sparse_fieldsets.Category')


class ArticleRouteSet(JsonapiRouteBase):
    model = apps.get_model('sparse_fieldsets.Article')

    fields: dict[str, SchemaField] = {
        None: SchemaFields(
            include={
                'author_detail': SchemaField(source='author_detail', required=False),
                'author_count': SchemaField(source='author_count', required=False),
                'some_count_property': SchemaField(source='some_count_property'),
            },
        ),
    }



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from constance.base import Config as ConfigBase


class Config(ConfigBase):
    """
    A class that extends the base configuration class from constance, providing
    custom attribute setting behavior.
    """

    def __setattr__(self, key, value):
        """
        Overrides the default __setattr__ method to set attributes on the configuration
        object, with exception handling to silently pass any errors.
        """
        try:
            super().__setattr__(key, value)
        except Exception:
            pass



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
For M2M relations, POST, PATCH, DELETE are tested both from the parent side and from the child side
"""

import json

import pytest
from bazis_test_utils.utils import get_api_client

from tests import factories


@pytest.mark.django_db(transaction=True)
def test_post_relationships_m2m_parent(sample_app):
    """
    We check setting up the relation from the parent side.
    We extend the relation with one existing record by adding two new ones
    """
    parent = factories.ParentEntityFactory.create(child_entities=False)
    child1 = factories.ChildEntityFactory.create()
    child2 = factories.ChildEntityFactory.create()
    child3 = factories.ChildEntityFactory.create()
    parent.child_entities.add(child1)

    payload = {
        'data': [
            {'type': 'entity.child_entity', 'id': str(child2.id)},
            {'type': 'entity.child_entity', 'id': str(child3.id)},
        ]
    }
    url = f'/api/v1/entity/parent_entity/{parent.id}/relationships/child_entities'
    response = get_api_client(sample_app).post(url, data=json.dumps(payload))
    assert response.status_code == 204
    parent.refresh_from_db()
    assert parent.child_entities.count() == 3
    assert parent.child_entities.filter(id=child1.id).exists()
    assert parent.child_entities.filter(id=child2.id).exists()
    assert parent.child_entities.filter(id=child3.id).exists()

    # we check passing an empty list. nothing changes.
    payload = {'data': []}
    response = get_api_client(sample_app).post(url, data=json.dumps(payload))
    assert response.status_code == 204
    parent.refresh_from_db()
    assert parent.child_entities.count() == 3


@pytest.mark.django_db(transaction=True)
def test_post_relationships_m2m_child(sample_app):
    """
    We check setting up the relation from the child side.
    We extend the relation with one existing record by adding two new ones
    """
    child = factories.ChildEntityFactory.create()
    parent1 = factories.ParentEntityFactory.create(child_entities=False)
    parent2 = factories.ParentEntityFactory.create(child_entities=False)
    parent3 = factories.ParentEntityFactory.create(child_entities=False)
    child.parent_entities.add(parent1)

    payload = {
        'data': [
            {'type': 'entity.parent_entity', 'id': str(parent2.id)},
            {'type': 'entity.parent_entity', 'id': str(parent3.id)},
        ]
    }
    url = f'/api/v1/entity/child_entity/{child.id}/relationships/parent_entities'
    response = get_api_client(sample_app).post(url, data=json.dumps(payload))
    assert response.status_code == 204
    child.refresh_from_db()
    assert child.parent_entities.count() == 3
    assert child.parent_entities.filter(id=parent1.id).exists()
    assert child.parent_entities.filter(id=parent2.id).exists()
    assert child.parent_entities.filter(id=parent3.id).exists()

    # we check passing an empty list. nothing changes.
    payload = {'data': []}
    response = get_api_client(sample_app).post(url, data=json.dumps(payload))
    assert response.status_code == 204
    child.refresh_from_db()
    assert child.parent_entities.count() == 3


@pytest.mark.django_db(transaction=True)
def test_patch_relationships_m2m_parent(sample_app):
    """
    We check replacing the relation from the parent side.
    We replace the relation with one existing record with a relation to two new ones
    """
    parent = factories.ParentEntityFactory.create(child_entities=False)
    child1 = factories.ChildEntityFactory.create()
    child2 = factories.ChildEntityFactory.create()
    child3 = factories.ChildEntityFactory.create()
    parent.child_entities.add(child1)

    payload = {
        'data': [
            {'type': 'entity.child_entity', 'id': str(child2.id)},
            {'type': 'entity.child_entity', 'id': str(child3.id)},
        ]
    }
    url = f'/api/v1/entity/parent_entity/{parent.id}/relationships/child_entities'
    response = get_api_client(sample_app).patch(url, data=json.dumps(payload))
    assert response.status_code == 204
    parent.refresh_from_db()
    assert parent.child_entities.count() == 2
    assert parent.child_entities.filter(id=child2.id).exists()
    assert parent.child_entities.filter(id=child3.id).exists()

    # we check passing an empty list. relations are cleared.
    payload = {'data': []}
    response = get_api_client(sample_app).patch(url, data=json.dumps(payload))
    assert response.status_code == 204
    parent.refresh_from_db()
    assert parent.child_entities.count() == 0


@pytest.mark.django_db(transaction=True)
def test_patch_relationships_m2m_child(sample_app):
    """
    We check replacing the relation from the child side.
    We replace the relation with one existing record with a relation to two new ones
    """
    child = factories.ChildEntityFactory.create()
    parent1 = factories.ParentEntityFactory.create()
    parent2 = factories.ParentEntityFactory.create()
    parent3 = factories.ParentEntityFactory.create()
    child.parent_entities.add(parent1)

    payload = {
        'data': [
            {'type': 'entity.parent_entity', 'id': str(parent2.id)},
            {'type': 'entity.parent_entity', 'id': str(parent3.id)},
        ]
    }
    url = f'/api/v1/entity/child_entity/{child.id}/relationships/parent_entities'
    response = get_api_client(sample_app).patch(url, data=json.dumps(payload))
    assert response.status_code == 204
    child.refresh_from_db()
    assert child.parent_entities.count() == 2
    assert child.parent_entities.filter(id=parent2.id).exists()
    assert child.parent_entities.filter(id=parent3.id).exists()

    # we check passing an empty list. relations are cleared.
    payload = {'data': []}
    response = get_api_client(sample_app).patch(url, data=json.dumps(payload))
    assert response.status_code == 204
    child.refresh_from_db()
    assert child.parent_entities.count() == 0


@pytest.mark.django_db(transaction=True)
def test_delete_relationships_m2m_parent(sample_app):
    """
    We check deleting the relation from the parent side.
    From two existing records we delete the one old record and keep the other
    """
    parent = factories.ParentEntityFactory.create(child_entities=False)
    child1 = factories.ChildEntityFactory.create()
    child2 = factories.ChildEntityFactory.create()
    parent.child_entities.add(child1, child2)

    payload = {'data': [{'type': 'entity.child_entity', 'id': str(child1.id)}]}
    url = f'/api/v1/entity/parent_entity/{parent.id}/relationships/child_entities'
    client = get_api_client(sample_app)

    # add the request method directly to the instance
    def request(method, url, **kwargs):
        headers = client.headers | kwargs.pop('headers', {})
        return client.client.request(method, url, headers=headers, **kwargs)

    client.request = request
    response = client.request('DELETE', url, json=payload)
    assert response.status_code == 204
    parent.refresh_from_db()
    assert parent.child_entities.count() == 1
    assert parent.child_entities.filter(id=child2.id).exists()

    # we check passing an empty list. nothing changes.
    payload = {'data': []}
    response = client.request('DELETE', url, json=payload)
    assert response.status_code == 204
    parent.refresh_from_db()
    assert parent.child_entities.count() == 1


@pytest.mark.django_db(transaction=True)
def test_delete_relationships_m2m_child(sample_app):
    """
    We check deleting the relation from the child side.
    From two existing records we delete the one old record and keep the other
    """
    child = factories.ChildEntityFactory.create()
    parent1 = factories.ParentEntityFactory.create()
    parent2 = factories.ParentEntityFactory.create()
    child.parent_entities.add(parent1, parent2)

    payload = {'data': [{'type': 'entity.parent_entity', 'id': str(parent1.id)}]}
    url = f'/api/v1/entity/child_entity/{child.id}/relationships/parent_entities'
    client = get_api_client(sample_app)

    # add the request method directly to the instance
    def request(method, url, **kwargs):
        headers = client.headers | kwargs.pop('headers', {})
        return client.client.request(method, url, headers=headers, **kwargs)

    client.request = request

    response = client.request('DELETE', url, json=payload)
    assert response.status_code == 204
    child.refresh_from_db()
    assert child.parent_entities.count() == 1
    assert child.parent_entities.filter(id=parent2.id).exists()

    # we check passing an empty list. nothing changes.
    payload = {'data': []}
    response = client.request('DELETE', url, json=payload)
    assert response.status_code == 204
    child.refresh_from_db()
    assert child.parent_entities.count() == 1



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing IsExists.
The following cases are checked:
case 1. One calculated IsExists field is added:
--- getting an item by id
--- getting a list
case 2. One calculated IsExists field is added with conditions on the fields of related tables specified in query via __:
--- getting an item by id
--- getting a list
case 3. An IsExists hierarchy is added with conditions on the fields of related tables specified in a lower‑level IsExists:
--- getting an item by id
--- getting a list

Cases 2 and 3 return the same result while forming different queries:
    case 2 - EXISTS with JOIN inside
    case 3 - EXISTS with nested EXISTS (subquery in a subquery)

Despite this, the PostgreSQL planner builds the same execution plan for the query.
Limit (rows=20 loops=1)
└─  Seq Scan on entity_vehicle as entity_vehicle (rows=20 loops=1)
    (sequential scan of rows of the entity_vehicle table with a limit of 20)
   └─ Gather (rows=447573 loops=1)
      (parallelization of the execution of the following subquery)
      └─ Hash Inner Join (rows=149191 loops=3)
         Hash Cond: (v0.driver_id = u0.id)
         ├─ Seq Scan on entity_carriertask as v0 (rows=149191 loops=3)
         │  (full scan of the entity_carriertask table with filters)
         │  Filter: ((dt_start IS NOT NULL) AND (dt_finish IS NULL))
         └─ Hash (rows=13087 loops=3)
            Buckets: 16384 Batches: 1 Memory Usage: 742 kB
             └─ Seq Scan on entity_driver (full scan of the entity_driver table with filters)
                 Filter: ((contact_phone IS NOT NULL) AND (((contact_phone)::text <> ''::text) OR (contact_phone IS NULL)))
Since the final query plan is the same, it is recommended to use
an IsExists hierarchy (case 3) to describe calculated IsExists fields, because it looks more readable than using "__" (case 2).

The query plan shown above also highlights that it is important not to forget about indexes.
The sequential Seq Scan over 149191 rows of entity_carriertask requires optimization.
Since the conditional cardinality of the combination of vehicle, dt_start and dt_finish is quite high, we add an index.
class CarrierTask(DtMixin, UuidMixin, JsonApiMixin):
    ...
    class Meta:
        ...
        indexes = [
            models.Index(fields=['vehicle', 'dt_start', 'dt_finish'], name='ix_vehicle_dates'),
        ]
Do not forget to warm up statistics with ANALYZE entity_carriertask;
New query plan.
Limit (rows=20 loops=1)
└─ Seq Scan on entity_vehicle (rows=20 loops=1)
   └─ Nested Loop Inner Join (rows=1 loops=20)
      ├─ Index Scan using ix_vehicle_dates on entity_carriertask u0 (rows=1 loops=20)
      │ Index Cond: (vehicle_id = entity_vehicle.id)
      │ Filter: (dt_start IS NOT NULL AND dt_finish IS NULL)
      └─ Index Scan using entity_driver_pkey on entity_driver u2 (rows=1 loops=19)
          Index Cond: (id = u0.driver_id)
          Filter: (contact_phone IS NOT NULL AND contact_phone <> '')
Since the new index provides high selectivity, the Seq Scan was successfully replaced with an Index Scan.
"""

import pytest
from bazis_test_utils.utils import get_api_client

from tests.utils.assert_sql import assert_sql_query, get_sql_query


@pytest.mark.django_db(transaction=True)
def test_isexists_one_field_item(sample_app, sample_vehicle_data):
    """Test for a single FieldIsExists when getting a single item by id and executing an SQL subquery
    @calc_property([
        FieldIsExists(
            source='carrier_tasks',
            query=Q(dt_start__isnull=False) & ~Q(dt_finish__isnull=False),
            alias='has_active_trip',
        ),
    ])
    """

    vehicle = sample_vehicle_data['vehicle']

    response = get_api_client(sample_app).get(
        f'/api/v1/has_active_trip/entity/vehicle/{str(vehicle.id)}'
    )

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               EXISTS
                  (SELECT 1 AS "a"
                   FROM "entity_carriertask" U0
                   WHERE (U0."vehicle_id" = ("entity_vehicle"."id")
                          AND U0."dt_start" IS NOT NULL
                          AND NOT (U0."dt_finish" IS NOT NULL))
                   LIMIT 1) AS "has_active_trip"
        FROM "entity_vehicle"
        WHERE "entity_vehicle"."id" = 'id_hex'::UUID
        ORDER BY "entity_vehicle"."id" ASC
        LIMIT 1;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    attrs = data['data']['attributes']

    assert 'has_active_trip' in attrs
    assert attrs['has_active_trip'] is True


@pytest.mark.django_db(transaction=True)
def test_isexists_one_field_list(sample_app, sample_vehicle_data):
    """Test for a single FieldIsExists when getting a single item by id and executing an SQL subquery
    @calc_property([
        FieldIsExists(
            source='carrier_tasks',
            query=Q(dt_start__isnull=False) & ~Q(dt_finish__isnull=False),
            alias='has_active_trip',
        ),
    ])
    """

    vehicle = sample_vehicle_data['vehicle']

    response = get_api_client(sample_app).get('/api/v1/has_active_trip/entity/vehicle/')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               EXISTS
                  (SELECT 1 AS "a"
                   FROM "entity_carriertask" U0
                   WHERE (U0."vehicle_id" = ("entity_vehicle"."id")
                          AND U0."dt_start" IS NOT NULL
                          AND NOT (U0."dt_finish" IS NOT NULL))
                   LIMIT 1) AS "has_active_trip"
        FROM "entity_vehicle"
        LIMIT 20;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    found = False
    for item in data['data']:
        if item['id'] != str(vehicle.id):
            continue
        found = True
        attrs = item['attributes']
        assert 'has_active_trip' in attrs
        assert attrs['has_active_trip'] is True
    assert found, f'Vehicle {vehicle.id} not found in response'


@pytest.mark.django_db(transaction=True)
def test_isexists_one_field_item_join(sample_app, sample_vehicle_data):
    """Test for a single FieldIsExists when getting an item by id and executing an SQL subquery with joins
    @calc_property([
        FieldIsExists(
            source='carrier_tasks',
            alias='has_active_trip_with_phone_join',
            query=Q(dt_start__isnull=False) & Q(dt_finish__isnull=True) & Q(driver__contact_phone__isnull=False)
                  & ~Q(driver__contact_phone='')
        )
    ])
    """

    vehicle = sample_vehicle_data['vehicle']

    response = get_api_client(sample_app).get(
        f'/api/v1/is_exists_join/entity/vehicle/{str(vehicle.id)}'
    )

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               EXISTS
                  (SELECT 1 AS "a"
                   FROM "entity_carriertask" U0
                   INNER JOIN "entity_driver" U2 ON (U0."driver_id" = U2."id")
                   WHERE (U0."vehicle_id" = ("entity_vehicle"."id")
                          AND U0."dt_start" IS NOT NULL
                          AND U0."dt_finish" IS NULL
                          AND U2."contact_phone" IS NOT NULL
                          AND NOT (U2."contact_phone" = ''
                                   AND U2."contact_phone" IS NOT NULL))
                   LIMIT 1) AS "has_active_trip_with_phone_join"
        FROM "entity_vehicle"
        WHERE "entity_vehicle"."id" = 'id_hex'::UUID
        ORDER BY "entity_vehicle"."id" ASC
        LIMIT 1;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    attrs = data['data']['attributes']

    assert 'has_active_trip_with_phone_join' in attrs
    assert attrs['has_active_trip_with_phone_join'] is True


@pytest.mark.django_db(transaction=True)
def test_isexists_one_field_list_join(sample_app, sample_vehicle_data):
    """Test for a single FieldIsExists when getting an item by id and executing an SQL subquery with joins
    @calc_property([
        FieldIsExists(
            source='carrier_tasks',
            alias='has_active_trip_with_phone_join',
            query=Q(dt_start__isnull=False) & Q(dt_finish__isnull=True) & Q(driver__contact_phone__isnull=False)
                  & ~Q(driver__contact_phone='')
        )
    ])
    """

    vehicle = sample_vehicle_data['vehicle']

    response = get_api_client(sample_app).get('/api/v1/is_exists_join/entity/vehicle/')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               EXISTS
          (SELECT 1 AS "a"
           FROM "entity_carriertask" U0
           INNER JOIN "entity_driver" U2 ON (U0."driver_id" = U2."id")
           WHERE (U0."vehicle_id" = ("entity_vehicle"."id")
                  AND U0."dt_start" IS NOT NULL
                  AND U0."dt_finish" IS NULL
                  AND U2."contact_phone" IS NOT NULL
                  AND NOT (U2."contact_phone" = ''
                           AND U2."contact_phone" IS NOT NULL))
           LIMIT 1) AS "has_active_trip_with_phone_join"
        FROM "entity_vehicle"
        LIMIT 20;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    found = False
    for item in data['data']:
        if item['id'] != str(vehicle.id):
            continue
        found = True
        attrs = item['attributes']
        assert 'has_active_trip_with_phone_join' in attrs
        assert attrs['has_active_trip_with_phone_join'] is True
    assert found, f'Vehicle {vehicle.id} not found in response'


@pytest.mark.django_db(transaction=True)
def test_exists_hierachy_item(sample_app, sample_vehicle_data):
    """Test for a FieldIsExists hierarchy when getting an item by id and executing an SQL subquery with subqueries
    @calc_property([
        FieldIsExists(
            source='carrier_tasks',
            alias='has_active_trip_with_phone_subquery',
            query=Q(dt_start__isnull=False) & Q(dt_finish__isnull=True) & Q(_has_drivers_with_phone=True),
            nested=[
                FieldIsExists(source='driver',
                              alias='_has_drivers_with_phone',
                              query=Q(contact_phone__isnull=False) & ~Q(contact_phone=''))
            ]
        )
    ])
    """

    vehicle = sample_vehicle_data['vehicle']

    response = get_api_client(sample_app).get(
        f'/api/v1/is_exists_subquery/entity/vehicle/{str(vehicle.id)}'
    )

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               EXISTS
                  (SELECT 1 AS "a"
                   FROM "entity_carriertask" V0
                   WHERE (V0."vehicle_id" = ("entity_vehicle"."id")
                          AND V0."dt_start" IS NOT NULL
                          AND V0."dt_finish" IS NULL
                          AND EXISTS
                            (SELECT 1 AS "a"
                             FROM "entity_driver" U0
                             WHERE (U0."id" = (V0."driver_id")
                                    AND U0."contact_phone" IS NOT NULL
                                    AND NOT (U0."contact_phone" = ''
                                             AND U0."contact_phone" IS NOT NULL))
                             LIMIT 1))
                   LIMIT 1) AS "has_active_trip_with_phone_subquery"
        FROM "entity_vehicle"
        WHERE "entity_vehicle"."id" = 'id_hex'::UUID
        ORDER BY "entity_vehicle"."id" ASC
        LIMIT 1;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    attrs = data['data']['attributes']

    assert 'has_active_trip_with_phone_subquery' in attrs
    assert attrs['has_active_trip_with_phone_subquery'] is True


@pytest.mark.django_db(transaction=True)
def test_exists_hierachy_list(sample_app, sample_vehicle_data):
    """Test for a FieldIsExists hierarchy when getting an item by id and executing an SQL subquery with subqueries
    @calc_property([
        FieldIsExists(
            source='carrier_tasks',
            alias='has_active_trip_with_phone_subquery',
            query=Q(dt_start__isnull=False) & Q(dt_finish__isnull=True) & Q(_has_drivers_with_phone=True),
            nested=[
                FieldIsExists(source='driver',
                              alias='_has_drivers_with_phone',
                              query=Q(contact_phone__isnull=False) & ~Q(contact_phone=''))
            ]
        )
    ])
    """

    vehicle = sample_vehicle_data['vehicle']

    response = get_api_client(sample_app).get('/api/v1/is_exists_subquery/entity/vehicle/')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               EXISTS
          (SELECT 1 AS "a"
           FROM "entity_carriertask" V0
           WHERE (V0."vehicle_id" = ("entity_vehicle"."id")
                  AND V0."dt_start" IS NOT NULL
                  AND V0."dt_finish" IS NULL
                  AND EXISTS
                    (SELECT 1 AS "a"
                     FROM "entity_driver" U0
                     WHERE (U0."id" = (V0."driver_id")
                            AND U0."contact_phone" IS NOT NULL
                            AND NOT (U0."contact_phone" = ''
                                     AND U0."contact_phone" IS NOT NULL))
                     LIMIT 1))
           LIMIT 1) AS "has_active_trip_with_phone_subquery"
        FROM "entity_vehicle"
        LIMIT 20;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    found = False
    for item in data['data']:
        if item['id'] != str(vehicle.id):
            continue
        found = True
        attrs = item['attributes']
        assert 'has_active_trip_with_phone_subquery' in attrs
        assert attrs['has_active_trip_with_phone_subquery'] is True
    assert found, f'Vehicle {vehicle.id} not found in response'



# Copyright 2026 EcoFuture Technology Services LLC and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from bazis_test_utils.utils import get_api_client

from tests.utils.assert_sql import assert_sql_query, get_sql_query


@pytest.mark.django_db(transaction=True)
def test_json_list(sample_app, sample_vehicle_data):
    """
        @calc_property([
        FieldJson(
            source='carrier_tasks',
            fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'driver_id'],
        ),
    ])
    """

    response = get_api_client(sample_app).get('/api/v1/json/entity/vehicle/')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               ARRAY
              (SELECT
                    JSONB_BUILD_OBJECT(('id')::text,
                    U0."id",
                    ('dt_start')::text,
                    U0."dt_start",
                    ('dt_finish')::text,
                    U0."dt_finish",
                    ('fact_waste_weight')::text,
                    U0."fact_waste_weight",
                    ('driver_id')::text,
                    U0."driver_id") AS "json"
              FROM "entity_carriertask" U0
              WHERE U0."vehicle_id" = ("entity_vehicle"."id")) AS "_carrier_tasks"
        FROM "entity_vehicle"
        LIMIT 20;"""

    assert_sql_query(expected_sql_template, sql_query)

    vehicle = sample_vehicle_data['vehicle']
    driver = sample_vehicle_data['driver']
    tasks = sample_vehicle_data['tasks']

    data = response.json()
    vehicles = data['data']
    found = False

    for v in vehicles:
        if v['id'] != str(vehicle.id):
            continue

        found = True
        attrs = v['attributes']

        assert 'carrier_task_list' in attrs
        task_list = attrs['carrier_task_list']
        assert len(task_list) == 3

        task_ids = [str(t.id) for t in tasks]
        weights = [float(t.fact_waste_weight) for t in tasks]

        for t in task_list:
            assert t['id'] in task_ids
            assert t['driver_id'] == str(driver.id)
            assert t['fact_waste_weight'] in weights

    assert found, f'Vehicle {vehicle.id} not found in response'


@pytest.mark.django_db(transaction=True)
def test_json_item(sample_app, sample_vehicle_data):
    """
        @calc_property([
        FieldJson(
            source='carrier_tasks',
            fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'driver_id'],
        ),
    ])
    """

    vehicle = sample_vehicle_data['vehicle']
    driver = sample_vehicle_data['driver']
    tasks = sample_vehicle_data['tasks']

    response = get_api_client(sample_app).get(f'/api/v1/json/entity/vehicle/{str(vehicle.id)}')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_vehicle"."dt_created",
               "entity_vehicle"."dt_updated",
               "entity_vehicle"."id",
               "entity_vehicle"."vehicle_model_id",
               "entity_vehicle"."country_id",
               "entity_vehicle"."gnum",
               ARRAY
              (SELECT
                    JSONB_BUILD_OBJECT(('id')::text,
                    U0."id",
                    ('dt_start')::text,
                    U0."dt_start",
                    ('dt_finish')::text,
                    U0."dt_finish",
                    ('fact_waste_weight')::text,
                    U0."fact_waste_weight",
                    ('driver_id')::text,
                    U0."driver_id") AS "json"
              FROM "entity_carriertask" U0
              WHERE U0."vehicle_id" = ("entity_vehicle"."id")) AS "_carrier_tasks"
        FROM "entity_vehicle"
        WHERE "entity_vehicle"."id" = 'id_hex'::UUID
        ORDER BY "entity_vehicle"."id" ASC
        LIMIT 1;"""

    assert_sql_query(expected_sql_template, sql_query, vehicle.id.hex)

    data = response.json()
    attrs = data['data']['attributes']

    assert 'carrier_task_list' in attrs
    task_list = attrs['carrier_task_list']
    assert len(task_list) == 3

    expected_tasks = {
        str(task.id): {
            'dt_start': task.dt_start.isoformat() if task.dt_start else None,
            'dt_finish': task.dt_finish.isoformat() if task.dt_finish else None,
            'fact_waste_weight': float(task.fact_waste_weight),
            'driver_id': str(driver.id),
        }
        for task in tasks
    }

    for t in task_list:
        assert t['id'] in expected_tasks
        expected = expected_tasks[t['id']]
        assert (t['dt_start'][:23] if t['dt_start'] else None) == (
            expected['dt_start'][:23] if expected['dt_start'] else None
        )
        assert (t['dt_finish'][:23] if t['dt_finish'] else None) == (
            expected['dt_finish'][:23] if expected['dt_finish'] else None
        )
        assert t['fact_waste_weight'] == expected['fact_waste_weight']
        assert t['driver_id'] == expected['driver_id']


@pytest.mark.django_db(transaction=True)
def test_json_hierarchy(sample_app, sample_vehicle_data):
    """
        @calc_property([
        FieldJson(
            source='carrier_tasks',
            fields=['id', 'dt_start', 'dt_finish', 'fact_waste_weight', 'vehicle_id'],
            nested=[
                FieldJson(source='vehicle',
                          fields=['gnum', 'dt_created'],
                          nested=[
                              FieldJson(source='vehicle_model',
                                        fields=['engine_type'])
                          ])]
        ),
    ])
    """

    response = get_api_client(sample_app).get('/api/v1/json/entity/driver/')

    assert response.status_code == 200

    sql_query = get_sql_query()
    expected_sql_template = """
        SELECT "entity_driver"."dt_created",
               "entity_driver"."dt_updated",
               "entity_driver"."id",
               "entity_driver"."first_name",
               "entity_driver"."last_name",
               "entity_driver"."contact_phone",
               "entity_driver"."org_owner_id",
               ARRAY
          (SELECT JSONB_BUILD_OBJECT(('id')::text, V0."id", ('_jsonapi_type')::text, 'entity.division') AS "json"
           FROM "entity_division" V0
           WHERE EXISTS
               (SELECT 1 AS "a"
                FROM "entity_driver_divisions" U0
                WHERE (U0."division_id" = (V0."id")
                       AND U0."driver_id" = ("entity_driver"."id"))
                LIMIT 1)) AS "divisions__ids",
               ARRAY
          (SELECT JSONB_BUILD_OBJECT(('id')::text, W0."id", ('dt_start')::text, W0."dt_start", ('dt_finish')::text, W0."dt_finish", ('fact_waste_weight')::text, W0."fact_waste_weight", ('vehicle_id')::text, W0."vehicle_id", ('_vehicle')::text, ARRAY
                                       (SELECT JSONB_BUILD_OBJECT(('gnum')::text, V0."gnum", ('dt_created')::text, V0."dt_created", ('_vehicle_model')::text, ARRAY
                                                                    (SELECT JSONB_BUILD_OBJECT(('engine_type')::text, U0."engine_type") AS "json"
                                                                     FROM "entity_vehiclemodel" U0
                                                                     WHERE U0."id" = (V0."vehicle_model_id"))) AS "json"
                                        FROM "entity_vehicle" V0
                                        WHERE V0."id" = (W0."vehicle_id"))) AS "json"
           FROM "entity_carriertask" W0
           WHERE W0."driver_id" = ("entity_driver"."id")) AS "_carrier_tasks"
        FROM "entity_driver"
        LIMIT 20;"""
    assert_sql_query(expected_sql_template, sql_query)

    driver = sample_vehicle_data['driver']
    vehicle = sample_vehicle_data['vehicle']
    tasks = sample_vehicle_data['tasks']

    data = response.json()
    driver_data = data['data'][0]
    attrs = driver_data['attributes']

    assert driver_data['id'] == str(driver.id)
    assert attrs['first_name'] == driver.first_name
    assert attrs['last_name'] == driver.last_name
    assert attrs['contact_phone'] == driver.contact_phone

    assert driver_data['relationships']['org_owner']['data']['id'] == str(
        sample_vehicle_data['organization'].id
    )
    assert driver_data['relationships']['org_owner']['data']['type'] == 'entity.organization'
    assert driver_data['relationships']['divisions']['data'] == []

    task_list = attrs['carrier_tasks_json_hierarchy']
    assert len(task_list) == 3

    for t in task_list:
        matching = [task for task in tasks if str(task.id) == t['id']]
        assert matching, f'Task {t["id"]} not found'
        task = matching[0]

        assert (t['dt_start'][:23] if t['dt_start'] else None) == (
            task.dt_start.isoformat()[:23] if task.dt_start else None
        )
        assert (t['dt_finish'][:23] if t['dt_finish'] else None) == (
            task.dt_finish.isoformat()[:23] if task.dt_finish else None
        )
        assert t['fact_waste_weight'] == float(task.fact_waste_weight)
        assert t['vehicle_gnum'] == vehicle.gnum
